#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Logger: handles logging functionalities."""


import logging
import os

from neural_compressor.common.utils import Mode

__all__ = [
    "level",
    "level_name",
    "Logger",  # TODO: not expose it
    "logger",
    "TuningLogger",
]


def _pretty_dict(value, indent=0):
    """Make the logger dict pretty."""
    prefix = "\n" + " " * (indent + 4)
    if isinstance(value, dict):
        items = [prefix + repr(key) + ": " + _pretty_dict(value[key], indent + 4) for key in value]
        return "{%s}" % (",".join(items) + "\n" + " " * indent)
    elif isinstance(value, list):
        items = [prefix + _pretty_dict(item, indent + 4) for item in value]
        return "[%s]" % (",".join(items) + "\n" + " " * indent)
    elif isinstance(value, tuple):
        items = [prefix + _pretty_dict(item, indent + 4) for item in value]
        return "(%s)" % (",".join(items) + "\n" + " " * indent)
    else:
        return repr(value)


class Logger(object):
    """Logger class."""

    __instance = None

    def __new__(cls):
        """Create a singleton Logger instance."""
        if Logger.__instance is None:
            Logger.__instance = object.__new__(cls)
            Logger.__instance._log()
        return Logger.__instance

    def _log(self):
        """Setup the logger format and handler."""
        LOGLEVEL = os.environ.get("LOGLEVEL", "INFO").upper()
        self._logger = logging.getLogger("neural_compressor")
        self._logger.handlers.clear()
        self._logger.setLevel(LOGLEVEL)
        formatter = logging.Formatter(
            "%(asctime)s [%(levelname)s][%(filename)s:%(lineno)d] %(message)s", "%Y-%m-%d %H:%M:%S"
        )
        streamHandler = logging.StreamHandler()
        streamHandler.setFormatter(formatter)
        self._logger.addHandler(streamHandler)
        self._logger.propagate = False

    def get_logger(self):
        """Get the logger."""
        return self._logger

    @staticmethod
    def log(level, msg, *args, **kwargs):
        """Output log with the level as a parameter."""
        kwargs.setdefault("stacklevel", 2)
        if isinstance(msg, dict):
            for _, line in enumerate(_pretty_dict(msg).split("\n")):
                Logger().get_logger().log(level, line, *args, **kwargs)
        else:
            Logger().get_logger().log(level, msg, *args, **kwargs)

    @staticmethod
    def debug(msg, *args, **kwargs):
        """Output log with the debug level."""
        kwargs.setdefault("stacklevel", 2)
        if isinstance(msg, dict):
            for _, line in enumerate(_pretty_dict(msg).split("\n")):
                Logger().get_logger().debug(line, *args, **kwargs)
        else:
            Logger().get_logger().debug(msg, *args, **kwargs)

    @staticmethod
    def error(msg, *args, **kwargs):
        """Output log with the error level."""
        kwargs.setdefault("stacklevel", 2)
        if isinstance(msg, dict):
            for _, line in enumerate(_pretty_dict(msg).split("\n")):
                Logger().get_logger().error(line, *args, **kwargs)
        else:
            Logger().get_logger().error(msg, *args, **kwargs)

    @staticmethod
    def fatal(msg, *args, **kwargs):
        """Output log with the fatal level."""
        kwargs.setdefault("stacklevel", 2)
        if isinstance(msg, dict):
            for _, line in enumerate(_pretty_dict(msg).split("\n")):
                Logger().get_logger().fatal(line, *args, **kwargs)
        else:
            Logger().get_logger().fatal(msg, *args, **kwargs)

    @staticmethod
    def info(msg, *args, **kwargs):
        """Output log with the info level."""
        kwargs.setdefault("stacklevel", 2)
        if isinstance(msg, dict):
            for _, line in enumerate(_pretty_dict(msg).split("\n")):
                Logger().get_logger().info(line, *args, **kwargs)
        else:
            Logger().get_logger().info(msg, *args, **kwargs)

    @staticmethod
    def warning(msg, *args, **kwargs):
        """Output log with the warning level (Alias of the method warn)."""
        kwargs.setdefault("stacklevel", 2)
        if isinstance(msg, dict):
            for _, line in enumerate(_pretty_dict(msg).split("\n")):
                Logger().get_logger().warning(line, *args, **kwargs)
        else:
            Logger().get_logger().warning(msg, *args, **kwargs)


level = Logger().get_logger().level
level_name = logging.getLevelName(level)

logger = Logger


def _get_log_msg(mode):
    log_msg = None
    if mode == Mode.QUANTIZE:
        log_msg = "Quantization"
    elif mode == Mode.PREPARE:  # pragma: no cover
        log_msg = "Preparation"
    elif mode == Mode.CONVERT:  # pragma: no cover
        log_msg = "Conversion"
    return log_msg


class TuningLogger:
    """A unified logger for the tuning/quantization process.

    It assists validation teams in retrieving logs.
    """

    @classmethod
    def tuning_start(cls) -> None:
        """Log the start of the tuning process."""
        logger.info("Tuning started.")

    @classmethod
    def trial_start(cls, trial_index: int = None) -> None:
        """Log the start of a trial."""
        logger.info("%d-trail started.", trial_index)

    @classmethod
    def execution_start(cls, mode=Mode.QUANTIZE, stacklevel=2):
        """Log the start of the execution process."""
        log_msg = _get_log_msg(mode)
        assert log_msg is not None, "Please check `mode` in execution_start function of TuningLogger class."
        logger.info("{} started.".format(log_msg), stacklevel=stacklevel)

    @classmethod
    def execution_end(cls, mode=Mode.QUANTIZE, stacklevel=2):
        """Log the end of the execution process."""
        log_msg = _get_log_msg(mode)
        assert log_msg is not None, "Please check `mode` in execution_end function of TuningLogger class."
        logger.info("{} end.".format(log_msg), stacklevel=stacklevel)

    @classmethod
    def evaluation_start(cls) -> None:
        """Log the start of the evaluation process."""
        logger.info("Evaluation started.")

    @classmethod
    def evaluation_end(cls) -> None:
        """Log the end of the evaluation process."""
        logger.info("Evaluation end.")

    @classmethod
    def trial_end(cls, trial_index: int = None) -> None:
        """Log the end of a trial."""
        logger.info("%d-trail end.", trial_index)

    @classmethod
    def tuning_end(cls) -> None:
        """Log the end of the tuning process."""
        logger.info("Tuning completed.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Share QDQ for ITEX Y pattern Graph Rewriter."""

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class ShareQDQForItexYPatternOptimizer(GraphRewriterBase):
    """Insert Q/DQ op before one input of Add to enable Conv/MatMul + BiasAdd + Add + Relu fusion for ITEX.

    Only 1 Q/DQ before Add op need to be inserted. Insert 2 Q/DQ breaks the ITEX fusion pattern.
    """

    @dump_elapsed_time("Pass ShareQDQForItexYPatternOptimizer")
    def do_transformation(self):
        """Share the QDQ of one output of Relu node with the another output which is Add node."""
        g = GraphAnalyzer()
        g.graph = self.model
        graph_info = g.parse_graph()

        patterns = [["Relu", "MaxPool"], ["QuantizeV2"], ["Dequantize"]]
        matched_nodes = g.query_fusion_pattern_nodes(patterns)

        for i in matched_nodes:
            relu_node_name = graph_info[i[0]].node.name
            if len(g.node_name_details[relu_node_name].outputs) != 2:
                continue

            add_node_name = g.node_name_details[relu_node_name].outputs[0]
            quantize_node_name = g.node_name_details[relu_node_name].outputs[1]
            if (
                "Add" not in g.node_name_details[add_node_name].node.op
                or g.node_name_details[quantize_node_name].node.op != "QuantizeV2"
            ):
                continue
            dequantize_node_name = graph_info[i[2]].node.name

            g.node_name_details[relu_node_name].outputs.remove(add_node_name)
            g.node_name_details[dequantize_node_name].outputs.append(add_node_name)
            g.node_name_details[add_node_name].node.input.remove(relu_node_name)
            g.node_name_details[add_node_name].node.input.append(dequantize_node_name)

        return g.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Split shared input Graph Rewriter."""

from tensorflow.core.framework import node_def_pb2

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class SplitSharedInputOptimizer(GraphRewriterBase):
    """Split the shared input if the input node is shared and const."""

    @dump_elapsed_time("Pass SplitSharedInputOptimizer")
    def do_transformation(self):
        """Execute splitting the shared input."""
        cur_graph = GraphAnalyzer()
        cur_graph.graph = self.model

        graph_info = cur_graph.parse_graph()

        is_shared_input = False
        # map of: input_name - op_name
        input_map = {}
        for node_name in list(graph_info.keys()):
            node = graph_info[node_name].node
            for _, input_node_name in enumerate(node.input):
                if input_node_name.startswith("^"):
                    continue
                if graph_info[Helper.node_name_from_input(input_node_name)].node.op == "Const":
                    # is shared and current node is not the first one
                    # sharing the input
                    if input_node_name in input_map:
                        is_shared_input = True
                        input_map[input_node_name].append(node.name)
                        new_input_node = node_def_pb2.NodeDef()
                        new_input_node.CopyFrom(graph_info[input_node_name].node)
                        new_input_node.name = input_node_name + "_nc_share_" + str(len(input_map[input_node_name]))
                        cur_graph.replace_const_node(new_input_node, [node.name], input_node_name, False)
                    else:
                        input_map[input_node_name] = [node.name]

        return cur_graph.dump_graph() if is_shared_input else self.model



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The tunable parameters module."""


import typing
from enum import Enum, auto
from typing import Any

from pydantic import BaseModel

from neural_compressor.common import logger


class ParamLevel(Enum):
    """Enumeration representing the different levels of tuning parameters.

    Attributes:
        OP_LEVEL: Represents the level of tuning parameters for operations.
        OP_TYPE_LEVEL: Represents the level of tuning parameters for operation types.
        MODEL_LEVEL: Represents the level of tuning parameters for models.
    """

    OP_LEVEL = auto()
    OP_TYPE_LEVEL = auto()
    MODEL_LEVEL = auto()


class TuningParam:
    """Define the tunable parameter for the algorithm.

    Example:
        Class FakeAlgoConfig(BaseConfig):
            '''Fake algo config.'''.

            params_list = [
                ...
                # For simple tunable types, like a list of int, giving
                # the param name is enough. `BaseConfig` class will
                # create the `TuningParam` implicitly.
                "simple_attr"

                # For complex tunable types, like a list of lists,
                # developers need to create the `TuningParam` explicitly.
                TuningParam("complex_attr", tunable_type=List[List[str]])

                # The default parameter level is `ParamLevel.OP_LEVEL`.
                # If the parameter is at a different level, developers need
                # to specify it explicitly.
                TuningParam("model_attr", level=ParamLevel.MODEL_LEVEL)

            ...

    # TODO: more examples to explain the usage of `TuningParam`.
    """

    def __init__(
        self,
        name: str,
        default_val: Any = None,
        tunable_type=None,
        options=None,
        level: ParamLevel = ParamLevel.OP_LEVEL,
    ) -> None:
        """Initialize a TuningParam object.

        Args:
            name (str): The name of the tuning parameter.
            default_val (Any, optional): The default value of the tuning parameter. Defaults to None.
            tunable_type (optional): The type of the tuning parameter. Defaults to None.
            options (optional): The available options for the tuning parameter. Defaults to None.
            level (ParamLevel, optional): The level of the tuning parameter. Defaults to ParamLevel.OP_LEVEL.
        """
        self.name = name
        self.default_val = default_val
        self.tunable_type = tunable_type
        self.options = options
        self.level = level

    @staticmethod
    def create_input_args_model(expect_args_type: Any):
        """Dynamically create an InputArgsModel based on the provided type hint.

        Args:
            expect_args_type (Any): The user-provided type hint for input_args.

        Returns:
            The dynamically created InputArgsModel class.
        """

        class DynamicInputArgsModel(BaseModel):
            input_args: expect_args_type

        return DynamicInputArgsModel

    def is_tunable(self, value: Any) -> bool:
        """Checks if the given value is tunable based on the specified tunable type.

        Args:
            value (Any): The value to be checked for tunability.

        Returns:
            bool: True if the value is tunable, False otherwise.
        """
        # Use `Pydantic` to validate the input_args.
        # TODO: refine the implementation in further.
        assert isinstance(
            self.tunable_type, typing._GenericAlias
        ), f"Expected a type hint, got {self.tunable_type} instead."
        DynamicInputArgsModel = TuningParam.create_input_args_model(self.tunable_type)
        try:
            new_args = DynamicInputArgsModel(input_args=value)
            return True
        except Exception as e:
            logger.debug(f"Failed to validate the input_args: {e}")
            return False

    def __str__(self) -> str:
        """Return the name of the tuning parameter."""
        return self.name



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intel Neural Compressor Tensorflow quantization AutoTune API."""


from copy import deepcopy
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import tensorflow as tf

from neural_compressor.common import logger
from neural_compressor.common.base_config import BaseConfig, get_all_config_set_from_config_registry
from neural_compressor.common.base_tuning import EvaluationFuncWrapper, TuningConfig, init_tuning
from neural_compressor.common.utils import call_counter, dump_elapsed_time
from neural_compressor.tensorflow.quantization import quantize_model
from neural_compressor.tensorflow.quantization.config import FRAMEWORK_NAME, StaticQuantConfig
from neural_compressor.tensorflow.utils import BaseModel, Model, constants

__all__ = [
    "autotune",
    "get_all_config_set",
]


def get_all_config_set() -> Union[BaseConfig, List[BaseConfig]]:
    """Get all config set."""
    return get_all_config_set_from_config_registry(fwk_name=FRAMEWORK_NAME)


@dump_elapsed_time("Pass auto-tune")
@call_counter
def autotune(
    model: Union[str, tf.keras.Model, BaseModel],
    tune_config: TuningConfig,
    eval_fn: Callable,
    eval_args: Optional[Tuple[Any]] = None,
    calib_dataloader: Callable = None,
    calib_iteration: int = 100,
    calib_func: Callable = None,
) -> Optional[BaseModel]:
    """The main entry of auto-tune."""
    model = Model(model)
    best_quant_model = None
    eval_func_wrapper = EvaluationFuncWrapper(eval_fn, eval_args)
    config_loader, tuning_logger, tuning_monitor = init_tuning(tuning_config=tune_config)
    baseline: float = eval_func_wrapper.evaluate(model)
    tuning_monitor.set_baseline(baseline)
    tuning_logger.tuning_start()
    for trial_index, quant_config in enumerate(config_loader, 1):
        tuning_logger.trial_start(trial_index=trial_index)
        tuning_logger.execution_start()
        logger.info(quant_config.to_dict())
        q_model = quantize_model(model, quant_config, calib_dataloader, calib_iteration, calib_func)
        tuning_logger.execution_end()
        tuning_logger.evaluation_start()
        eval_result: float = eval_func_wrapper.evaluate(q_model)
        tuning_logger.evaluation_end()
        tuning_monitor.add_trial_result(trial_index, eval_result, quant_config)
        tuning_logger.trial_end(trial_index)
        if tuning_monitor.need_stop():
            logger.info("Stopped tuning.")
            best_trial_record = tuning_monitor.get_best_trial_record()
            if best_trial_record.trial_index != trial_index:
                logger.info("Re-quantizing with best quantization config...")
                del q_model
                best_quant_config: BaseConfig = best_trial_record.quant_config
                best_quant_model = quantize_model(
                    model, best_quant_config, calib_dataloader, calib_iteration, calib_func
                )
            else:
                best_quant_model = q_model
            break
    tuning_logger.tuning_end()
    return best_quant_model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Initialize custom depthwise conv2d layers for Keras quantization."""

import json

import tensorflow as tf
from tensorflow.keras import activations, constraints, initializers, regularizers

from neural_compressor.tensorflow.utils import version1_gte_version2

if version1_gte_version2(tf.__version__, "2.16.1"):
    from keras.src import ops
    from keras.src.layers.convolutional.base_depthwise_conv import BaseDepthwiseConv  # pylint: disable=E0401
elif version1_gte_version2(tf.__version__, "2.13.0"):
    from keras.src.layers.convolutional.base_depthwise_conv import DepthwiseConv  # pylint: disable=E0401
    from keras.src.utils import conv_utils, tf_utils  # pylint: disable=E0401
else:
    from keras.layers.convolutional.base_depthwise_conv import DepthwiseConv  # pylint: disable=E0401
    from keras.utils import conv_utils, tf_utils  # pylint: disable=E0401

if version1_gte_version2(tf.__version__, "2.16.1"):

    class QDepthwiseConv2D(BaseDepthwiseConv):  # pragma: no cover
        """The custom quantized DepthwiseConv2D layer."""

        def __init__(
            self,
            kernel_size,
            strides=(1, 1),
            padding="valid",
            depth_multiplier=1,
            data_format=None,
            dilation_rate=(1, 1),
            activation=None,
            use_bias=True,
            depthwise_initializer="glorot_uniform",
            bias_initializer="zeros",
            depthwise_regularizer=None,
            bias_regularizer=None,
            activity_regularizer=None,
            depthwise_constraint=None,
            bias_constraint=None,
            act_min_value=None,
            act_max_value=None,
            weight_min_value=None,
            weight_max_value=None,
            granularity="per_tensor",
            quant_status="calib",
            quant_mode="SCALED",
            quant_T="s8",
            quant_round_mode="HALF_AWAY_FROM_ZERO",
            quant_narrow_range=False,
            quant_axis=None,
            **kwargs
        ):
            """Initialize custom quantized DepthwiseConv2D layer."""
            super().__init__(
                2,
                kernel_size=kernel_size,
                strides=strides,
                padding=padding,
                depth_multiplier=depth_multiplier,
                data_format=data_format,
                dilation_rate=dilation_rate,
                activation=activation,
                use_bias=use_bias,
                depthwise_initializer=depthwise_initializer,
                bias_initializer=bias_initializer,
                depthwise_regularizer=depthwise_regularizer,
                bias_regularizer=bias_regularizer,
                activity_regularizer=activity_regularizer,
                depthwise_constraint=depthwise_constraint,
                bias_constraint=bias_constraint,
                **kwargs
            )
            T_map = {"s8": tf.qint8, "u8": tf.quint8}
            self.weight_min_value = weight_min_value
            self.weight_max_value = weight_max_value
            self.act_min_value = act_min_value
            self.act_max_value = act_max_value
            self.granularity = granularity
            self.quant_status = quant_status
            self.quant_mode = quant_mode
            self.quant_T = T_map[quant_T]
            self.quant_round_mode = quant_round_mode
            self.quant_narrow_range = quant_narrow_range
            self.quant_axis = quant_axis

        def call(self, inputs):
            """The __call__ function of custom quantized DepthwiseConv2D layer."""
            if self.quant_status == "calib" and not isinstance(inputs, tf.keras.KerasTensor):
                if self.granularity == "per_tensor":
                    self.act_min_value = tf.math.reduce_min(inputs)
                    self.act_max_value = tf.math.reduce_max(inputs)
                else:
                    self.act_min_value = tf.math.reduce_min(inputs, axis=1)
                    self.act_max_value = tf.math.reduce_max(inputs, axis=1)
                kernel = self.kernel
            elif self.quant_status == "quantize":
                assert (
                    self.act_min_value is not None
                ), "Invalid activation min-max values, please check calibration process"
                inputs, _, _ = tf.quantization.quantize(
                    inputs,
                    self.act_min_value,
                    self.act_max_value,
                    self.quant_T,
                    mode=self.quant_mode,
                    round_mode=self.quant_round_mode,
                    narrow_range=self.quant_narrow_range,
                    axis=self.quant_axis,
                )
                inputs = tf.quantization.dequantize(
                    inputs,
                    self.act_min_value,
                    self.act_max_value,
                    mode=self.quant_mode,
                    narrow_range=self.quant_narrow_range,
                    axis=self.quant_axis,
                )

                # add the Q/DQ here
                kernel, _, _ = tf.quantization.quantize(
                    self.kernel, self.weight_min_value, self.weight_max_value, tf.qint8, axis=3, mode="SCALED"
                )
                kernel = tf.quantization.dequantize(
                    kernel,
                    self.weight_min_value,
                    self.weight_max_value,
                    axis=3,
                    mode="SCALED",
                )

            input_channel = self._get_input_channel(inputs.shape)
            outputs = ops.depthwise_conv(
                inputs,
                kernel,
                strides=self.strides,
                padding=self.padding,
                dilation_rate=self.dilation_rate,
                data_format=self.data_format,
            )

            if self.use_bias:
                if self.data_format == "channels_last":
                    bias_shape = (1,) * (self.rank + 1) + (self.depth_multiplier * input_channel,)
                else:
                    bias_shape = (1, self.depth_multiplier * input_channel) + (1,) * self.rank
                bias = ops.reshape(self.bias, bias_shape)
                outputs += bias

            if self.activation is not None:
                return self.activation(outputs)
            return outputs

        @classmethod
        def from_config(cls, config):
            """Deserialize this class from a config dict."""
            return cls(**config)

        def get_config(self):
            """Serialize this class to a config dict."""
            config = super(QDepthwiseConv2D, self).get_config()
            config.update(
                {
                    "act_min_value": self.act_min_value,
                    "act_max_value": self.act_max_value,
                    "weight_min_value": self.weight_min_value,
                    "weight_max_value": self.weight_max_value,
                    "granularity": self.granularity,
                    "quant_status": self.quant_status,
                    "quant_mode": self.quant_mode,
                    "quant_T": "s8" if self.quant_T == tf.qint8 else "u8",
                    "quant_round_mode": self.quant_round_mode,
                    "quant_narrow_range": self.quant_narrow_range,
                    "quant_axis": self.quant_axis,
                }
            )

            return config

else:

    class QDepthwiseConv2D(DepthwiseConv):
        """The custom quantized DepthwiseConv2D layer."""

        def __init__(
            self,
            kernel_size,
            strides=(1, 1),
            padding="valid",
            depth_multiplier=1,
            data_format=None,
            dilation_rate=(1, 1),
            activation=None,
            use_bias=True,
            depthwise_initializer="glorot_uniform",
            bias_initializer="zeros",
            depthwise_regularizer=None,
            bias_regularizer=None,
            activity_regularizer=None,
            depthwise_constraint=None,
            bias_constraint=None,
            act_min_value=None,
            act_max_value=None,
            weight_min_value=None,
            weight_max_value=None,
            granularity="per_tensor",
            quant_status="calib",
            quant_mode="SCALED",
            quant_T="s8",
            quant_round_mode="HALF_AWAY_FROM_ZERO",
            quant_narrow_range=False,
            quant_axis=None,
            **kwargs
        ):
            """Initialize custom quantized DepthwiseConv2D layer."""
            super().__init__(
                2,
                kernel_size=kernel_size,
                strides=strides,
                padding=padding,
                depth_multiplier=depth_multiplier,
                data_format=data_format,
                dilation_rate=dilation_rate,
                activation=activation,
                use_bias=use_bias,
                depthwise_initializer=depthwise_initializer,
                bias_initializer=bias_initializer,
                depthwise_regularizer=depthwise_regularizer,
                bias_regularizer=bias_regularizer,
                activity_regularizer=activity_regularizer,
                depthwise_constraint=depthwise_constraint,
                bias_constraint=bias_constraint,
                **kwargs
            )
            T_map = {"s8": tf.qint8, "u8": tf.quint8}
            self.weight_min_value = weight_min_value
            self.weight_max_value = weight_max_value
            self.act_min_value = act_min_value
            self.act_max_value = act_max_value
            self.granularity = granularity
            self.quant_status = quant_status
            self.quant_mode = quant_mode
            self.quant_T = T_map[quant_T]
            self.quant_round_mode = quant_round_mode
            self.quant_narrow_range = quant_narrow_range
            self.quant_axis = quant_axis

        def call(self, inputs):
            """The __call__ function of custom quantized DepthwiseConv2D layer."""
            if self.quant_status == "calib":
                if self.granularity == "per_tensor":
                    self.act_min_value = tf.math.reduce_min(inputs)
                    self.act_max_value = tf.math.reduce_max(inputs)
                else:
                    self.act_min_value = tf.math.reduce_min(inputs, axis=1)
                    self.act_max_value = tf.math.reduce_max(inputs, axis=1)
                depthwise_kernel = self.depthwise_kernel
            elif self.quant_status == "quantize":
                assert (
                    self.act_min_value is not None
                ), "Invalid activation min-max values, please check calibration process"
                inputs, _, _ = tf.quantization.quantize(
                    inputs,
                    self.act_min_value,
                    self.act_max_value,
                    self.quant_T,
                    mode=self.quant_mode,
                    round_mode=self.quant_round_mode,
                    narrow_range=self.quant_narrow_range,
                    axis=self.quant_axis,
                )
                inputs = tf.quantization.dequantize(
                    inputs,
                    self.act_min_value,
                    self.act_max_value,
                    mode=self.quant_mode,
                    narrow_range=self.quant_narrow_range,
                    axis=self.quant_axis,
                )

                # add the Q/DQ here
                depthwise_kernel, _, _ = tf.quantization.quantize(
                    self.depthwise_kernel, self.weight_min_value, self.weight_max_value, tf.qint8, axis=3, mode="SCALED"
                )
                depthwise_kernel = tf.quantization.dequantize(
                    depthwise_kernel,
                    self.weight_min_value,
                    self.weight_max_value,
                    axis=3,
                    mode="SCALED",
                )

            outputs = tf.keras.backend.depthwise_conv2d(
                inputs,
                depthwise_kernel,
                strides=self.strides,
                padding=self.padding,
                data_format=self.data_format,
                dilation_rate=self.dilation_rate,
            )

            if self.use_bias:
                outputs = tf.keras.backend.bias_add(outputs, self.bias, data_format=self.data_format)

            if self.activation is not None:
                return self.activation(outputs)

            return outputs

        @classmethod
        def from_config(cls, config):
            """Deserialize this class from a config dict."""
            return cls(**config)

        def get_config(self):
            """Serialize this class to a config dict."""
            config = super(QDepthwiseConv2D, self).get_config()
            config.update(
                {
                    "act_min_value": self.act_min_value,
                    "act_max_value": self.act_max_value,
                    "weight_min_value": self.weight_min_value,
                    "weight_max_value": self.weight_max_value,
                    "granularity": self.granularity,
                    "quant_status": self.quant_status,
                    "quant_mode": self.quant_mode,
                    "quant_T": "s8" if self.quant_T == tf.qint8 else "u8",
                    "quant_round_mode": self.quant_round_mode,
                    "quant_narrow_range": self.quant_narrow_range,
                    "quant_axis": self.quant_axis,
                }
            )

            return config

        @tf_utils.shape_type_conversion
        def compute_output_shape(self, input_shape):
            """Compute_output_shape."""
            if self.data_format == "channels_first":
                rows = input_shape[2]
                cols = input_shape[3]
                out_filters = input_shape[1] * self.depth_multiplier
            elif self.data_format == "channels_last":
                rows = input_shape[1]
                cols = input_shape[2]
                out_filters = input_shape[3] * self.depth_multiplier

            rows = conv_utils.conv_output_length(
                rows,
                self.kernel_size[0],
                self.padding,
                self.strides[0],
                self.dilation_rate[0],
            )
            cols = conv_utils.conv_output_length(
                cols,
                self.kernel_size[1],
                self.padding,
                self.strides[1],
                self.dilation_rate[1],
            )
            if self.data_format == "channels_first":
                return (input_shape[0], out_filters, rows, cols)
            elif self.data_format == "channels_last":
                return (input_shape[0], rows, cols, out_filters)


def initialize_int8_depthwise_conv2d(fp32_layer, q_config):
    """Initialize int8 depthwise conv2d."""
    kwargs = fp32_layer.get_config()
    q_name = fp32_layer.name

    param_list = [
        "name",
        "kernel_size",
        "strides",
        "padding",
        "depth_multiplier",
        "data_format",
        "dilation_rate",
        "activation",
        "use_bias",
        "depthwise_initializer",
        "bias_initializer",
        "depthwise_regularizer",
        "activity_regularizer",
        "bias_regularizer",
        "depthwise_constraint",
        "bias_constraint",
    ]
    for p in param_list:  # pragma: no cover
        if p in kwargs:
            del kwargs[p]

    return QDepthwiseConv2D(
        name=q_name,
        kernel_size=fp32_layer.kernel_size,
        strides=fp32_layer.strides,
        padding=fp32_layer.padding,
        depth_multiplier=fp32_layer.depth_multiplier,
        data_format=fp32_layer.data_format,
        dilation_rate=fp32_layer.dilation_rate,
        activation=fp32_layer.activation,
        use_bias=fp32_layer.use_bias,
        depthwise_initializer=fp32_layer.depthwise_initializer,
        bias_initializer=fp32_layer.bias_initializer,
        depthwise_regularizer=fp32_layer.depthwise_regularizer,
        bias_regularizer=fp32_layer.bias_regularizer,
        activity_regularizer=fp32_layer.activity_regularizer,
        depthwise_constraint=fp32_layer.depthwise_constraint,
        bias_constraint=fp32_layer.bias_constraint,
        quant_T=q_config["T"],
        granularity=q_config["granularity"],
        **kwargs
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The base config."""

from __future__ import annotations

import inspect
import json
import os
import re
from abc import ABC, abstractmethod
from collections import OrderedDict
from itertools import product
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

from typing_extensions import Self

from neural_compressor.common.tuning_param import TuningParam
from neural_compressor.common.utils import (
    BASE_CONFIG,
    COMPOSABLE_CONFIG,
    DEFAULT_WHITE_LIST,
    DEFAULT_WORKSPACE,
    EMPTY_WHITE_LIST,
    GLOBAL,
    LOCAL,
    OP_NAME_OR_MODULE_TYPE,
    logger,
)

__all__ = [
    "options",
    "register_config",
    "get_all_config_set_from_config_registry",
    "register_supported_configs_for_fwk",
    "BaseConfig",
    "ConfigRegistry",
    "ComposableConfig",
]


class ConfigRegistry(object):
    """A registry for managing configuration classes for different algorithms within specific frameworks."""

    registered_configs = {}
    _config_registry = None

    def __new__(cls) -> Self:
        """Create a new instance of the ConfigRegistry class.

        This method is responsible for creating a new instance of the ConfigRegistry class.
        It ensures that only one instance of the class is created by checking if the `_config_registry`
        attribute is None. If it is None, a new instance is created and assigned to `_config_registry`.
        If `_config_registry` is not None, the existing instance is returned.

        Returns:
            The instance of the ConfigRegistry class.
        """
        if cls._config_registry is None:
            cls._config_registry = super(ConfigRegistry, cls).__new__(cls)

        return cls._config_registry

    @classmethod
    def register_config_impl(cls, framework_name: str, algo_name: str, priority: Union[float, int] = 0):
        """Register a configuration decorator.

        This decorator is used to register the configuration classes
        for different algorithms within specific frameworks.

        Usage example:
            @ConfigRegistry.register_config_impl(framework_name=FRAMEWORK_NAME, algo_name=ExampleAlgorithm, priority=1)
            class ExampleAlgorithmConfig:
                # Configuration details for the ExampleAlgorithm

        Args:
            framework_name (str): The framework name.
            algo_name (str): The algorithm name.
            priority (Union[float, int], optional): The priority of the configuration.
                A larger number indicates a higher priority, which will be tried first
                at the auto-tune stage. Defaults to 0.
        """

        def decorator(config_cls):
            cls.registered_configs.setdefault(framework_name, {})
            cls.registered_configs[framework_name][algo_name] = {"priority": priority, "cls": config_cls}
            return config_cls

        return decorator

    @classmethod
    def get_all_configs(cls) -> Dict[str, Dict[str, Dict[str, object]]]:
        """Get all registered configurations.

        Returns:
            Dict[str, Dict[str, Dict[str, object]]]: A dictionary containing all registered configurations.
        """
        return cls.registered_configs

    @classmethod
    def get_sorted_configs(cls) -> Dict[str, OrderedDict[str, Dict[str, object]]]:
        """Get registered configurations sorted by priority.

        Returns:
            Dict[str, OrderedDict[str, Dict[str, object]]]:
                A dictionary containing registered configurations sorted by priority.
        """
        sorted_configs = OrderedDict()
        for framework_name, algos in sorted(cls.registered_configs.items()):
            sorted_configs[framework_name] = OrderedDict(
                sorted(algos.items(), key=lambda x: x[1]["priority"], reverse=True)
            )
        return sorted_configs

    @classmethod
    def get_cls_configs(cls) -> Dict[str, Dict[str, object]]:
        """Get registered configurations without priority.

        Returns:
            Dict[str, Dict[str, object]]: A dictionary containing registered configurations without priority.
        """
        cls_configs = {}
        for framework_name, algos in cls.registered_configs.items():
            cls_configs[framework_name] = {}
            for algo_name, config_data in algos.items():
                cls_configs[framework_name][algo_name] = config_data["cls"]
        return cls_configs

    @classmethod
    def get_all_config_cls_by_fwk_name(cls, fwk_name: str) -> List[Type[BaseConfig]]:
        """Get all registered configuration classes for a specific framework.

        Args:
            fwk_name (str): The framework name.

        Returns:
            List[Type[BaseConfig]]: A list of all registered configuration classes for the specified framework.
        """
        configs_cls = []
        for algo_name, config_pairs in cls.registered_configs.get(fwk_name, {}).items():
            configs_cls.append(config_pairs["cls"])
        return configs_cls


config_registry = ConfigRegistry()


def register_config(framework_name: str, algo_name: str, priority: Union[float, int] = 0):
    """Register config decorator.

    The register the configuration classes for different algorithms within specific frameworks.

    Usage example:
        @register_config(framework_name=FRAMEWORK_NAME, algo_name=ExampleAlgorithm, priority=100)
        class ExampleAlgorithmConfig:
            # Configuration details for the ExampleAlgorithm

    Args:
        framework_name: the framework name.
        algo_name: the algorithm name.
        priority: the priority of the configuration. A larger number indicates a higher priority,
            which will be tried first at the auto-tune stage. Defaults to 0.
    """
    return config_registry.register_config_impl(framework_name=framework_name, algo_name=algo_name, priority=priority)


class BaseConfig(ABC):
    """The base config for all algorithm configs.

    Attributes:
        name (str): The name of the config.
        params_list (list): The list of **tunable parameters** in the config.
    """

    name = BASE_CONFIG
    params_list = []

    def __init__(self, white_list: Optional[List[OP_NAME_OR_MODULE_TYPE]] = DEFAULT_WHITE_LIST) -> None:
        """Initialize the BaseConfig.

        Args:
            white_list (Optional[List[OP_NAME_OR_MODULE_TYPE]]): The white list of operator names or types.
                Defaults to DEFAULT_WHITE_LIST.
        """
        self._global_config: Optional[BaseConfig] = None
        # For PyTorch, operator_type is the collective name for module type and functional operation type,
        # for example, `torch.nn.Linear`, and `torch.nn.functional.linear`.
        # local config is the collections of operator_type configs and operator configs.
        self._local_config: Dict[str, Optional[BaseConfig]] = {}
        self._white_list = white_list

    def _post_init(self):
        if self.white_list == DEFAULT_WHITE_LIST:
            global_config = self.get_params_dict()
            self._global_config = self.__class__(**global_config, white_list=None)
        elif isinstance(self.white_list, list) and len(self.white_list) > 0:
            for op_name_or_type in self.white_list:
                global_config = self.get_params_dict()
                tmp_config = self.__class__(**global_config, white_list=None)
                self.set_local(op_name_or_type, tmp_config)
        elif self.white_list == EMPTY_WHITE_LIST:
            return
        else:
            raise NotImplementedError(
                f"The white list should be one of {DEFAULT_WHITE_LIST}, {EMPTY_WHITE_LIST},"
                " a not empty list, but got {self.white_list}"
            )

    @property
    def white_list(self):
        """Get the white list of operator names or types.

        Returns:
            The white list of operator names or types.
        """
        return self._white_list

    @white_list.setter
    def white_list(self, op_name_or_type_list: Optional[List[OP_NAME_OR_MODULE_TYPE]]):
        """Set the white list of operator names or types.

        Args:
            op_name_or_type_list (Optional[List[OP_NAME_OR_MODULE_TYPE]]): The white list of operator names or types.
        """
        self._white_list = op_name_or_type_list

    @property
    def global_config(self):
        """Get the global configuration object.

        Returns:
            The global configuration object.
        """
        return self._global_config

    @global_config.setter
    def global_config(self, config):
        """Set the global configuration object.

        Args:
            config: The global configuration object.
        """
        self._global_config = config

    @property
    def local_config(self):
        """Get the local configuration objects.

        Returns:
            The local configuration objects.
        """
        return self._local_config

    @local_config.setter
    def local_config(self, config):
        """Set the local configuration objects.

        Args:
            config: The local configuration objects.
        """
        self._local_config = config

    def set_local(self, operator_name_or_list: Union[List, str, Callable], config: BaseConfig) -> BaseConfig:
        """Set custom configuration based on the global configuration object.

        Args:
            operator_name_or_list (Union[List, str, Callable]): Specific operator name or list of operator names.
            config (BaseConfig): Specific configuration.

        Returns:
            Updated Config.
        """
        if isinstance(operator_name_or_list, list):
            for operator_name in operator_name_or_list:
                if operator_name in self.local_config:
                    logger.warning("The configuration for %s has already been set, update it.", operator_name)
                self.local_config[operator_name] = config
        else:
            if operator_name_or_list in self.local_config:
                logger.warning("The configuration for %s has already been set, update it.", operator_name_or_list)
            self.local_config[operator_name_or_list] = config
        return self

    def to_dict(self):
        """Convert the config to a dictionary.

        Returns:
            The config as a dictionary.
        """
        result = {}
        global_config = self.get_params_dict()
        if bool(self.local_config):
            result[LOCAL] = {}
            for op_name, config in self.local_config.items():
                result[LOCAL][op_name] = config.to_dict()
            if self.global_config:
                result[GLOBAL] = global_config
        else:
            result = global_config
        return result

    def get_params_dict(self):
        """Get a dictionary containing the parameters and their values for the current instance.

        Returns:
            A dictionary containing the parameters and their values.
        """
        result = dict()
        for param, value in self.__dict__.items():
            if param not in ["_global_config", "_local_config", "_white_list"]:
                result[param] = value
        return result

    @classmethod
    def from_dict(cls, config_dict):
        """Construct config from a dictionary.

        Args:
            config_dict: The dictionary containing the config.

        Returns:
            The constructed config.
        """
        if GLOBAL not in config_dict and LOCAL not in config_dict:
            config = cls(**config_dict)
            return config
        else:
            config = cls(**config_dict.get(GLOBAL, {}))
            operator_config = config_dict.get(LOCAL, {})
            if operator_config:
                for op_name, op_config in operator_config.items():
                    config.set_local(op_name, cls(**op_config))
            return config

    @classmethod
    def to_diff_dict(cls, instance) -> Dict[str, Any]:
        """Compare the instance with the default BaseConfig and return the differences as a dictionary.

        Args:
            instance: The instance to compare.

        Returns:
            A dictionary representation of the instance with only the differences from the class defaults.
        """
        # TODO (Yi) to implement it
        return {}

    @classmethod
    def from_json_file(cls, filename):
        """Load config from a JSON file.

        Args:
            filename (str): The path to the JSON file.

        Returns:
            The loaded config.
        """
        with open(filename, "r", encoding="utf-8") as file:
            config_dict = json.load(file)
        return cls.from_dict(**config_dict)

    def to_json_file(self, filename):
        """Save the config to a JSON file.

        Args:
            filename (str): The path to save the JSON file.
        """
        config_dict = self.to_dict()
        with open(filename, "w", encoding="utf-8") as file:
            json.dump(config_dict, file, indent=4)
        logger.info("Dump the config into %s.", filename)

    def to_json_string(self, use_diff: bool = False) -> str:
        """Serializes this instance to a JSON string.

        Args:
            use_diff (bool, optional): If True, only the difference between the config instance and the default
                BaseConfig is serialized to JSON string. Defaults to False.

        Returns:
            The config as a JSON string.
        """
        if use_diff is True:
            config_dict = self.to_diff_dict(self)
        else:
            config_dict = self.to_dict()
        try:
            return json.dumps(config_dict, indent=2) + "\n"
        except:
            return config_dict

    def __repr__(self) -> str:
        """Return a string representation of the config.

        Returns:
            str: The string representation of the config.
        """
        return f"{self.__class__.__name__} {self.to_json_string()}"

    @classmethod
    @abstractmethod
    def register_supported_configs(cls):
        """Add all supported configs."""
        raise NotImplementedError

    @classmethod
    def validate(self, user_config: BaseConfig):
        """Validates the user configuration.

        Args:
            user_config (BaseConfig): The user configuration to be validated.

        Returns:
            None
        """
        # TODO(Yi) validate the user config
        pass

    def __add__(self, other: BaseConfig) -> BaseConfig:
        """Combine two configs.

        If the other config is an instance of the same class, the local configs will be combined.
        Otherwise, a `ComposableConfig` will be created to combine the two configs.

        Args:
            other (BaseConfig): The other config to combine.

        Returns:
            BaseConfig: The combined config.
        """
        if isinstance(other, type(self)):
            for op_name, config in other.local_config.items():
                self.set_local(op_name, config)
            return self
        else:
            return ComposableConfig(configs=[self, other])

    @staticmethod
    def get_the_default_value_of_param(config: BaseConfig, param: str) -> Any:
        """Get the default value of a parameter in the config.

        Args:
            config (BaseConfig): The config object.
            param (str): The name of the parameter.

        Returns:
            default_vaule: The default value of the parameter.
        """
        # Get the signature of the __init__ method
        signature = inspect.signature(config.__init__)

        # Get the parameters and their default values
        parameters = signature.parameters
        return parameters.get(param).default

    def expand(self) -> List[BaseConfig]:
        """Expand the config.

        case 1
            {
                "global": { "weight_bits": [4, 6]}
            }
            expand to :
            1st trial config:
            {
                "global": { "weight_bits": 4}
            }
            2nd trial config:
            {
                "global": { "weight_bits": 6}
            }
        case 2
        # TODO (Yi) to support the expansion of config with `local`
        {
            "global": {
                "weight_bits": [4, 6]
            },
            "local":
            {
                "fc1":{
                    "weight_bits": [6, 8]
                },
                "fc2":{
                    "weight_bits": [4]
                }
            }

        } -> ?
        """
        config_list: List[BaseConfig] = []
        params_list = self.params_list
        config = self
        tuning_param_list = []
        not_tuning_param_pair = {}  # key is the param name, value is the user specified value
        for param in params_list:
            # Create `TuningParam` for each param
            # There are two cases:
            # 1. The param is a string.
            # 2. The param is a `TuningParam` instance.
            if isinstance(param, str):
                default_param = self.get_the_default_value_of_param(config, param)
                tuning_param = TuningParam(name=param, tunable_type=List[type(default_param)])
            elif isinstance(param, TuningParam):
                tuning_param = param
            else:
                raise ValueError(f"Unsupported param type: {param}")
            # Assign the options to the `TuningParam` instance
            param_val = getattr(config, tuning_param.name)
            if param_val is not None:
                if tuning_param.is_tunable(param_val):
                    tuning_param.options = param_val
                    tuning_param_list.append(tuning_param)
                else:
                    not_tuning_param_pair[tuning_param.name] = param_val
        logger.debug("Tuning param list: %s", tuning_param_list)
        logger.debug("Not tuning param pair: %s", not_tuning_param_pair)
        if len(tuning_param_list) == 0:
            config_list = [config]
        else:
            # The `TuningParam` instance with no options will cause the product to be empty.
            # Filter out the `TuningParam` instances with no options
            tuning_param_list = list(filter(lambda x: len(x.options) > 0, tuning_param_list))
            tuning_param_name_lst = [tuning_param.name for tuning_param in tuning_param_list]
            for params_values in product(*[tuning_param.options for tuning_param in tuning_param_list]):
                tuning_param_pair = dict(zip(tuning_param_name_lst, params_values))
                tmp_params_dict = {**not_tuning_param_pair, **tuning_param_pair}
                new_config = self.__class__(**tmp_params_dict)
                logger.info(new_config.to_dict())
                config_list.append(new_config)
        logger.info("Expanded the %s and got %d configs.", self.__class__.name, len(config_list))
        return config_list

    def _get_op_name_op_type_config(self):
        op_type_config_dict = dict()
        op_name_config_dict = dict()
        for name, config in self.local_config.items():
            if self._is_op_type(name):
                # Convert the Callable to String.
                new_name = self._op_type_to_str(name)
                op_type_config_dict[new_name] = config
            else:
                op_name_config_dict[name] = config
        return op_type_config_dict, op_name_config_dict

    def to_config_mapping(
        self, config_list: List[BaseConfig] = None, model_info: List[Tuple[str, str]] = None
    ) -> OrderedDict[Union[str, str], OrderedDict[str, BaseConfig]]:
        """Generate the configuration mapping based on the model information.

        Args:
            config_list (List[BaseConfig], optional): A list of BaseConfig objects to be converted.
                If not provided, the method will use the current instance of BaseConfig. Defaults to None.
            model_info (List[Tuple[str, str]], optional): A list of tuples representing the model information.
                Each tuple contains the operation name and operation type. Defaults to None.

        Returns:
            OrderedDict[Union[str, str], OrderedDict[str, BaseConfig]]:
                A OrderedDict representing the configuration mapping.
                The keys of the outer OrderedDict are tuples of (operation name, operation type),
                and the values are inner OrderedDicts containing the corresponding configuration objects.
        """
        config_mapping = OrderedDict()
        if config_list is None:
            config_list = [self]
        for config in config_list:
            global_config = config.global_config
            op_type_config_dict, op_name_config_dict = config._get_op_name_op_type_config()
            for op_name, op_type in model_info:
                if self.global_config is not None:
                    config_mapping[(op_name, op_type)] = global_config
                if op_type in op_type_config_dict:
                    config_mapping[(op_name, op_type)] = op_type_config_dict[op_type]
                for op_name_pattern in op_name_config_dict:
                    if re.match(op_name_pattern, op_name):
                        config_mapping[(op_name, op_type)] = op_name_config_dict[op_name_pattern]
        return config_mapping

    @staticmethod
    def _op_type_to_str(op_type: Callable) -> str:
        # * Ort and TF may override this method.
        op_type_name = getattr(op_type, "__name__", "")
        if op_type_name == "":
            logger.warning("The op_type %s has no attribute __name__.", op_type)
        return op_type_name

    @staticmethod
    def _is_op_type(name: str) -> bool:
        # * Ort and TF may override this method.
        return not isinstance(name, str)

    @classmethod
    @abstractmethod
    def get_config_set_for_tuning(cls):
        """A set of predefined configurations used for tuning.

        This method should be implemented by subclasses to provide a set of configurations
        that can be used for auto-tune.

        Returns:
            set: A set of configurations for tuning.

        Raises:
            NotImplementedError: If the method is not implemented by the subclass.
        """
        raise NotImplementedError

    def __eq__(self, other: BaseConfig) -> bool:
        """Check if the current BaseConfig object is equal to another BaseConfig object.

        Args:
            other (BaseConfig): The other BaseConfig object to compare with.

        Returns:
            bool: True if the objects are equal, False otherwise.
        """
        if not isinstance(other, type(self)):
            return False
        return self.params_list == other.params_list and all(
            getattr(self, str(attr)) == getattr(other, str(attr)) for attr in self.params_list
        )


class ComposableConfig(BaseConfig):
    """A class representing a composable configuration.

    This class allows for composing multiple configurations together by using the `+` operator.

    Args:
        configs (List[BaseConfig]): A list of base configurations to be composed.

    Attributes:
        config_list (List[BaseConfig]): The list of base configurations.
    """

    name = COMPOSABLE_CONFIG

    def __init__(self, configs: List[BaseConfig]) -> None:
        """Initializes a new ComposableConfig.

        Args:
            configs (List[BaseConfig]): A list of BaseConfig objects.

        Returns:
            None
        """
        self.config_list = configs

    def __add__(self, other: BaseConfig) -> BaseConfig:
        """Adds another BaseConfig object to the current BaseConfig object.

        If the other object is of the same type as the current object, the config_list of the other object is appended
        to the config_list of the current object. Otherwise, the other object is appended directly to the config_list.

        Args:
            other (BaseConfig): The other BaseConfig object to be added.

        Returns:
            BaseConfig: The updated BaseConfig object after the addition.
        """
        if isinstance(other, type(self)):
            self.config_list.extend(other.config_list)
        else:
            self.config_list.append(other)
        return self

    def to_dict(self, params_list=[], operator2str=None):
        """Converts the configuration object to a dictionary.

        Args:
            params_list (list): A list of parameters to include in the dictionary.
                If empty, all parameters will be included.
            operator2str (callable): A function that converts operator objects to strings.
                If None, the default conversion will be used.

        Returns:
            dict: A dictionary representation of the configuration object.
        """
        result = {}
        for config in self.config_list:
            result[config.name] = config.to_dict()
        return result

    @classmethod
    def from_dict(cls, config_dict: OrderedDict[str, Dict], config_registry: Dict[str, BaseConfig]):
        """Create a BaseConfig object from a dictionary representation.

        Args:
            config_dict (OrderedDict[str, Dict]): The dictionary representation of the configuration.
            config_registry (Dict[str, BaseConfig]): The registry of available configurations.

        Returns:
            BaseConfig: The created BaseConfig object.

        Raises:
            AssertionError: If the config_dict does not include at least one configuration.
        """
        assert len(config_dict) >= 1, "The config dict must include at least one configuration."
        num_configs = len(config_dict)
        name, value = next(iter(config_dict.items()))
        config = config_registry[name].from_dict(value)
        for _ in range(num_configs - 1):
            name, value = next(iter(config_dict.items()))
            config += config_registry[name].from_dict(value)
        return config

    def to_json_string(self, use_diff: bool = False) -> str:
        """Convert the object to a JSON string representation.

        Args:
            use_diff (bool): Whether to include only the differences from the base configuration.
                Defaults to False.

        Returns:
            str: The JSON string representation of the object.
        """
        return json.dumps(self.to_dict(), indent=2) + "\n"

    def __repr__(self) -> str:
        """A string representation of the object.

        Returns:
            str: The string representation of the object.
        """
        return f"{self.__class__.__name__} {self.to_json_string()}"

    def to_config_mapping(
        self, config_list: List[BaseConfig] = None, model_info: Dict[str, Any] = None
    ) -> OrderedDict[str, BaseConfig]:
        """Converts the configuration list to a mapping of (op_name, op_type) to corresponding BaseConfig objects.

        Args:
            config_list (List[BaseConfig], optional): List of BaseConfig objects. Defaults to None.
            model_info (Dict[str, Any], optional): Dictionary containing model information. Defaults to None.

        Returns:
            OrderedDict[str, BaseConfig]: Mapping of (op_name, op_type) to corresponding BaseConfig objects.
        """
        config_mapping = OrderedDict()
        for config in self.config_list:
            op_type_config_dict, op_name_config_dict = config._get_op_name_op_type_config()
            single_config_model_info = model_info.get(config.name, None)
            for op_name, op_type in single_config_model_info:
                if op_type in op_type_config_dict:
                    config_mapping[(op_name, op_type)] = op_name_config_dict[op_type]
                for op_name_pattern in op_name_config_dict:
                    if re.match(op_name_pattern, op_name):
                        config_mapping[(op_name, op_type)] = op_name_config_dict[op_name_pattern]
        return config_mapping

    @classmethod
    def register_supported_configs(cls):
        """Add all supported configs."""
        raise NotImplementedError

    @classmethod
    def get_config_set_for_tuning(cls) -> None:
        """Get the set of predefined configurations used for tuning."""
        # TODO (Yi) handle the composable config in `tuning_config`
        return None

    def get_model_info(self, model, *args, **kwargs):
        """Get the model information."""
        model_info_dict = dict()
        for config in self.config_list:
            model_info_dict.update({config.name: config.get_model_info(model, *args, **kwargs)})
        return model_info_dict


def get_all_config_set_from_config_registry(fwk_name: str) -> Union[BaseConfig, List[BaseConfig]]:
    """Retrieves all the configuration sets from the config registry for a given framework name.

    Args:
        fwk_name (str): The name of the framework.

    Returns:
        Union[BaseConfig, List[BaseConfig]]: The configuration set(s) for the given framework name.
    """
    all_registered_config_cls: List[BaseConfig] = config_registry.get_all_config_cls_by_fwk_name(fwk_name)
    config_set = []
    for config_cls in all_registered_config_cls:
        config_set.append(config_cls.get_config_set_for_tuning())
    return config_set


def register_supported_configs_for_fwk(fwk_name: str):
    """Register supported configs for specific framework.

    Args:
        fwk_name: the framework name.
    """
    all_registered_config_cls: List[BaseConfig] = config_registry.get_all_config_cls_by_fwk_name(fwk_name)
    for config_cls in all_registered_config_cls:
        config_cls.register_supported_configs()


#######################################################
####   Options
#######################################################


def _check_value(name, src, supported_type, supported_value=[]):
    """Check if the given object is the given supported type and in the given supported value.

    Example::

        from neural_compressor.common.base_config import _check_value

        def datatype(self, datatype):
            if _check_value("datatype", datatype, list, ["fp32", "bf16", "uint8", "int8"]):
                self._datatype = datatype
    """
    if isinstance(src, list) and any([not isinstance(i, supported_type) for i in src]):
        assert False, "Type of {} items should be {} but not {}".format(
            name, str(supported_type), [type(i) for i in src]
        )
    elif not isinstance(src, list) and not isinstance(src, supported_type):
        assert False, "Type of {} should be {} but not {}".format(name, str(supported_type), type(src))

    if len(supported_value) > 0:
        if isinstance(src, str) and src not in supported_value:
            assert False, "{} is not in supported {}: {}. Skip setting it.".format(src, name, str(supported_value))
        elif (
            isinstance(src, list)
            and all([isinstance(i, str) for i in src])
            and any([i not in supported_value for i in src])
        ):
            assert False, "{} is not in supported {}: {}. Skip setting it.".format(src, name, str(supported_value))

    return True


class Options:
    """Option Class for configs.

    This class is used for configuring global variables. The global variable options is created with this class.
    If you want to change global variables, you should use functions from neural_compressor.common.utils.utility.py:
        set_random_seed(seed: int)
        set_workspace(workspace: str)
        set_resume_from(resume_from: str)
        set_tensorboard(tensorboard: bool)

    Args:
        random_seed(int): Random seed used in neural compressor.
                          Default value is 1978.
        workspace(str): The directory where intermediate files and tuning history file are stored.
                        Default value is:
                            "./nc_workspace/{}/".format(datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")).
        resume_from(str): The directory you want to resume tuning history file from.
                          The tuning history was automatically saved in the workspace directory
                               during the last tune process.
                          Default value is None.
        tensorboard(bool): This flag indicates whether to save the weights of the model and the inputs of each layer
                               for visual display.
                           Default value is False.

    Example::

        from neural_compressor.common import set_random_seed, set_workspace, set_resume_from, set_tensorboard
        set_random_seed(2022)
        set_workspace("workspace_path")
        set_resume_from("workspace_path")
        set_tensorboard(True)
    """

    def __init__(self, random_seed=1978, workspace=DEFAULT_WORKSPACE, resume_from=None, tensorboard=False):
        """Init an Option object."""
        os.makedirs(workspace, exist_ok=True)
        self.random_seed = random_seed
        self.workspace = workspace
        self.resume_from = resume_from
        self.tensorboard = tensorboard

    @property
    def random_seed(self):
        """Get random seed."""
        return self._random_seed

    @random_seed.setter
    def random_seed(self, random_seed):
        """Set random seed."""
        if _check_value("random_seed", random_seed, int):
            self._random_seed = random_seed

    @property
    def workspace(self):
        """Get workspace."""
        os.makedirs(self._workspace, exist_ok=True)
        return self._workspace

    @workspace.setter
    def workspace(self, workspace):
        """Set workspace."""
        if _check_value("workspace", workspace, str):
            self._workspace = workspace

    @property
    def resume_from(self):
        """Get resume_from."""
        return self._resume_from

    @resume_from.setter
    def resume_from(self, resume_from):
        """Set resume_from."""
        if resume_from is None or _check_value("resume_from", resume_from, str):
            self._resume_from = resume_from

    @property
    def tensorboard(self):
        """Get tensorboard."""
        return self._tensorboard

    @tensorboard.setter
    def tensorboard(self, tensorboard):
        """Set tensorboard."""
        if _check_value("tensorboard", tensorboard, bool):
            self._tensorboard = tensorboard


options = Options()



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intel Neural Compressor Tensorflow quantization base API."""


from typing import Any, Callable, Dict, Tuple, Union

import tensorflow as tf

from neural_compressor.common import logger
from neural_compressor.common.base_config import BaseConfig, ComposableConfig, config_registry
from neural_compressor.common.utils import STATIC_QUANT, log_process
from neural_compressor.tensorflow.utils import BaseModel, KerasModel, Model, algos_mapping


def need_apply(configs_mapping: Dict[Tuple[str, callable], BaseConfig], algo_name):
    """Whether to apply the algorithm."""
    return any(config.name == algo_name for config in configs_mapping.values())


@log_process()
def quantize_model(
    model: Union[str, tf.keras.Model, BaseModel],
    quant_config: Union[BaseConfig, list],
    calib_dataloader: Callable = None,
    calib_iteration: int = 100,
    calib_func: Callable = None,
):
    """The main entry to quantize model.

    Args:
        model: a fp32 model to be quantized.
        quant_config: single or lists of quantization configuration.
        calib_dataloader: a data loader for calibration.
        calib_iteration: the iteration of calibration.
        calib_func: the function used for calibration, should be a substitution for calib_dataloader
        when the built-in calibration function of INC does not work for model inference.

    Returns:
        q_model: the quantized model.
    """
    q_model = Model(model)
    if isinstance(quant_config, list):
        for config in quant_config:
            q_model = quantize_model_with_single_config(q_model, config, calib_dataloader, calib_iteration, calib_func)
    else:
        q_model = quantize_model_with_single_config(
            q_model, quant_config, calib_dataloader, calib_iteration, calib_func
        )

    return q_model


def quantize_model_with_single_config(
    q_model: BaseModel,
    quant_config: BaseConfig,
    calib_dataloader: Callable = None,
    calib_iteration: int = 100,
    calib_func: Callable = None,
):
    """Quantize model using single config.

    Args:
        model: a model wrapped by INC TF model class.
        quant_config: a quantization configuration.
        calib_dataloader: a data loader for calibration.
        calib_iteration: the iteration of calibration.
        calib_func: the function used for calibration, should be a substitution for calib_dataloader
        when the built-in calibration function of INC does not work for model inference.

    Returns:
        q_model: the quantized model.
    """
    framework_name = "keras" if isinstance(q_model, KerasModel) else "tensorflow"
    registered_configs = config_registry.get_cls_configs()
    if isinstance(quant_config, dict):
        quant_config = ComposableConfig.from_dict(quant_config, config_registry=registered_configs[framework_name])
        logger.info(f"Parsed a config dict to construct the quantization config: {quant_config}.")
    else:
        assert isinstance(
            quant_config, BaseConfig
        ), f"Please pass a dict or config instance as the quantization configuration, but got {type(quant_config)}."
    logger.info(f"Quantize model with config: \n {quant_config.to_json_string()} \n")
    # select quantization algo according to config

    model_info = quant_config.get_model_info(model=q_model)
    configs_mapping = quant_config.to_config_mapping(model_info=model_info)
    logger.debug(configs_mapping)
    for algo_name, algo_func in algos_mapping.items():
        if need_apply(configs_mapping, algo_name):
            logger.info(f"Start to apply {algo_name} on the model.")
            q_model = algo_func(q_model, configs_mapping, calib_dataloader, calib_iteration, calib_func)
    return q_model



#
#  -*- coding: utf-8 -*-
#
#  Copyright (c) 2021 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""Graph Converter Class."""

import copy
import logging
import os
import tempfile
from collections import OrderedDict, UserDict

import tensorflow as tf
from tensorflow.core.framework import graph_pb2
from tensorflow.python.platform import gfile

from neural_compressor.tensorflow.quantization.utils.graph_rewriter.bf16.bf16_convert import BF16Convert
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.generic.fold_batch_norm import (
    FoldBatchNormNodesOptimizer,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.generic.fuse_pad_with_conv import (
    FusePadWithConv2DOptimizer,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.generic.fuse_pad_with_fp32_conv import (
    FusePadWithFP32Conv2DOptimizer,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.generic.insert_print_node import (
    InsertPrintMinMaxNode,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.generic.remove_training_nodes import (
    RemoveTrainingNodesOptimizer,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.generic.strip_equivalent_nodes import (
    StripEquivalentNodesOptimizer,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.generic.strip_unused_nodes import (
    StripUnusedNodesOptimizer,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.int8.freeze_fake_quant import (
    FreezeFakeQuantOpOptimizer,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.int8.freeze_value import FreezeValueTransformer
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.int8.fuse_conv_redundant_dequantize import (
    FuseConvRedundantDequantizeTransformer,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.int8.fuse_conv_requantize import (
    FuseConvRequantizeTransformer,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.int8.fuse_matmul_redundant_dequantize import (
    FuseMatMulRedundantDequantizeTransformer,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.int8.fuse_matmul_requantize import (
    FuseMatMulRequantizeDequantizeNewAPITransformer,
    FuseMatMulRequantizeDequantizeTransformer,
    FuseMatMulRequantizeNewAPITransformer,
    FuseMatMulRequantizeTransformer,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.int8.meta_op_optimizer import (
    MetaInfoChangingMemOpOptimizer,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.int8.post_hostconst_converter import (
    PostHostConstConverter,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.int8.post_quantized_op_cse import PostCseOptimizer
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.int8.scale_propagation import (
    ScaleProPagationTransformer,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.qdq.insert_qdq_pattern import (
    GenerateGraphWithQDQPattern,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.qdq.merge_duplicated_qdq import (
    MergeDuplicatedQDQOptimizer,
)
from neural_compressor.tensorflow.quantization.utils.graph_rewriter.qdq.share_qdq_y_pattern import (
    ShareQDQForItexYPatternOptimizer,
)
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.quantization.utils.quantize_graph.qdq.optimize_qdq import OptimizeQDQGraph
from neural_compressor.tensorflow.quantization.utils.quantize_graph.quantize_graph_for_intel_cpu import (
    QuantizeGraphForIntel,
)
from neural_compressor.tensorflow.quantization.utils.quantize_graph_common import QuantizeGraphHelper
from neural_compressor.tensorflow.quantization.utils.transform_graph.bias_correction import BiasCorrection
from neural_compressor.tensorflow.quantization.utils.transform_graph.insert_logging import InsertLogging
from neural_compressor.tensorflow.quantization.utils.transform_graph.rerange_quantized_concat import (
    RerangeQuantizedConcat,
)
from neural_compressor.tensorflow.quantization.utils.utility import generate_feed_dict, iterator_sess_run
from neural_compressor.tensorflow.utils import (
    SPR_BASE_VERSIONS,
    CaptureOutputToFile,
    CpuInfo,
    Model,
    TensorflowSavedModelModel,
    combine_histogram,
    deep_get,
    get_all_fp32_data,
    get_tensor_histogram,
    version1_eq_version2,
    version1_gt_version2,
    version1_gte_version2,
    version1_lt_version2,
    version1_lte_version2,
)

TF_SUPPORTED_MAX_VERSION = "2.16.1"
TF_SUPPORTED_MIN_VERSION = "1.14.0"

logger = logging.getLogger("neural_compressor")
debug = bool(logger.level == logging.DEBUG)


class GraphConverter:
    """Graph Converter Class is used to generate the quantization graph."""

    def __init__(
        self,
        model,
        qt_config={},
        recipes={},
        int8_sequences={},
        fp32_ops=[],
        bf16_ops=[],
        data_loader=None,
        calib_func=None,
        fake_quant=False,
        itex_mode=False,
        qdq_enabled=False,
        new_api=False,
        performance_only=False,
        use_bf16=False,
    ):
        """Convert graph.

        :param model: input tensorflow model.
        :param qt_config: quantization configs, including iteration and op-wise quant config
        :param fp32_ops: fall back to fp32 dtype op list
        :param bf16_ops: fall back to bf16 dtype op list
        :param data_loader: for calibration phase used dataloader
        :param calib_func: for calibration phase used function
        :param fake_quant: for quantization-aware training model conversion to default model
        """
        self.model = model
        # (TODO) does it right to make the internal model format as graph_def
        self.output_tensor_names = self.model.output_tensor_names
        self.input_tensor_names = self.model.input_tensor_names
        # quantize specific config
        self.calib_iteration = qt_config["calib_iteration"] if not fake_quant else 0
        self.op_wise_config = qt_config["op_wise_config"]
        self.advance_config = deep_get(qt_config, "advance")
        self.device = qt_config["device"] if "device" in qt_config else "cpu"
        self.int8_sequences = int8_sequences
        self.fp32_ops = fp32_ops
        self.bf16_ops = bf16_ops
        self.recipes = recipes
        self.fake_quant = fake_quant
        self.itex_mode = itex_mode
        self.qdq_enabled = qdq_enabled
        self.performance_only = performance_only
        self.quantized_node_info = []
        self._calibration_data = []
        self._fp32_print_data = []
        self.data_loader = data_loader
        self.calib_func = calib_func
        self._check_tf_version()
        self._check_args()

        if "backend" in self.model.kwargs:
            self._fp32_model = Model(self.model._model, **self.model.kwargs)
        else:
            self._fp32_model = Model(
                self.model._model,
                **self.model.kwargs,
                backend="itex" if itex_mode and not isinstance(self.model, TensorflowSavedModelModel) else "default"
            )
        self._fp32_model.graph_def = self.model.graph_def
        self._fp32_model.output_tensor_names = self.output_tensor_names
        self._fp32_model.input_tensor_names = self.input_tensor_names

        self._gen_tmp_filenames()
        self._kl_op_dict = {}
        self._kl_keys = []
        self._llm_weight_minmax = {}
        self._print_node_mapping = {}
        self._enable_kl_op_names = [k for k in self.op_wise_config if self.op_wise_config[k][1] == "kl"]
        self.scale_info = {}
        self.scale_info.update(qt_config)
        self.scale_info.update({"recipes": self.recipes})
        self.scale_info.update({"int8_sequences": self.int8_sequences})
        self.scale_info.update({"bf16_ops": self.bf16_ops})
        self.scale_info.update({"fp32_ops": self.fp32_ops})

        if "backend" in self.model.kwargs:  # pragma: no cover
            self._sampling_model = Model(self.model._model, **self.model.kwargs)
        else:
            self._sampling_model = Model(
                self.model._model,
                **self.model.kwargs,
                backend="itex" if itex_mode and not isinstance(self.model, TensorflowSavedModelModel) else "default"
            )
        self._sampling_model.output_tensor_names = self.output_tensor_names
        self._sampling_model.input_tensor_names = self.input_tensor_names

        if self.performance_only:
            # reuse the fp32 model for performance only mode
            self._tmp_graph_def = self.model.graph_def
        else:
            self._tmp_graph_def = copy.deepcopy(self.model.graph_def)
        self.new_api = new_api  # bool(version1_gte_version2(tf.version.VERSION, '2.8.0'))
        self.use_bf16 = use_bf16
        self.exclude_node_names = []

    # pylint: disable=no-member
    def _inference(self, model):
        """Run the calibration on the input graph.

        Args:
            model(TensorflowBaseModel): input TensorflowBaseModel
        """
        if self.calib_func:
            self.calib_func(model)
            return

        if model.model_type == "llm_saved_model":
            self._inference_llm(model)
            return

        # ITEX optimization has broken INC calibration process.
        # INC needs turn off ITEX optimization pass in calibration stage.
        # TODO ITEX will provide API to replace setting environment variable.
        os.environ["ITEX_REMAPPER"] = "0"
        sess = model.sess
        iter_op = model.iter_op
        input_tensor = model.input_tensor
        output_tensor = model.output_tensor
        # TF table initialization: https://github.com/tensorflow/tensorflow/issues/8665
        node_names = [node.name for node in sess.graph.as_graph_def().node]
        if "init_all_tables" in node_names:  # pragma: no cover
            init_table_op = sess.graph.get_operation_by_name("init_all_tables")
            sess.run(init_table_op)

        logger.info("Start sampling on calibration dataset.")
        if hasattr(self.data_loader, "__len__") and len(self.data_loader) == 0:  # pragma: no cover
            feed_dict = {}
            _ = (
                sess.run(output_tensor, feed_dict)
                if iter_op == []
                else iterator_sess_run(sess, iter_op, feed_dict, output_tensor, self.calib_iteration)
            )
        for idx, (inputs, labels) in enumerate(self.data_loader):
            if len(input_tensor) == 1:
                feed_dict = {}
                if (
                    isinstance(inputs, dict) or isinstance(inputs, OrderedDict) or isinstance(inputs, UserDict)
                ):  # pragma: no cover
                    for name in inputs:
                        for tensor in input_tensor:
                            pos = tensor.name.rfind(":")
                            t_name = tensor.name if pos < 0 else tensor.name[:pos]
                            if name == t_name:
                                feed_dict[tensor] = inputs[name]
                                break
                else:
                    feed_dict = {input_tensor[0]: inputs}  # get raw tensor using index [0]
            else:  # pragma: no cover
                assert len(input_tensor) == len(inputs), "inputs len must equal with input_tensor"
                feed_dict = {}
                if isinstance(inputs, dict) or isinstance(inputs, OrderedDict) or isinstance(inputs, UserDict):
                    for name in inputs:
                        for tensor in input_tensor:
                            pos = tensor.name.rfind(":")
                            t_name = tensor.name if pos < 0 else tensor.name[:pos]
                            if name in [tensor.name, t_name]:
                                feed_dict[tensor] = inputs[name]
                                break
                else:
                    # sometimes the input_tensor is not the same order with inputs
                    # we should check and pair them
                    def check_shape(tensor, data):
                        # scalar or 1 dim default True
                        if (
                            tensor.shape is None
                            or tensor.shape == tf.TensorShape(None)
                            or len(tensor.shape.dims) == 1
                            or not hasattr(data, "shape")
                        ):
                            return True
                        tensor_shape = tuple(tensor.shape)
                        data_shape = tuple(data.shape)
                        for tensor_dim, data_dim in zip(tensor_shape, data_shape):
                            if tensor_dim is not None and tensor_dim != data_dim:
                                return False
                        return True

                    disorder_tensors = []
                    disorder_inputs = []
                    for idx, sort_tensor in enumerate(input_tensor):
                        sort_input = inputs[idx]
                        if check_shape(sort_tensor, sort_input):
                            feed_dict.update({sort_tensor: sort_input})
                        else:
                            disorder_tensors.append(sort_tensor)
                            disorder_inputs.append(sort_input)
                    for i, dis_tensor in enumerate(disorder_tensors):
                        for j, dis_input in enumerate(disorder_inputs):
                            if check_shape(dis_tensor, dis_input):
                                feed_dict.update({dis_tensor: dis_input})
                                break
            _ = (
                sess.run(output_tensor, feed_dict)
                if iter_op == []
                else iterator_sess_run(sess, iter_op, feed_dict, output_tensor, self.calib_iteration)
            )
            if idx + 1 == self.calib_iteration:
                break
        os.environ["ITEX_REMAPPER"] = "1"

    def _inference_llm(self, model):
        input_tensor_names = model.input_tensor_names
        auto_trackable = model.model
        infer = auto_trackable.signatures["serving_default"]
        for idx, (inputs, _) in enumerate(self.data_loader):
            feed_dict = {}
            if len(input_tensor_names) == 1:
                feed_dict[input_tensor_names[0]] = inputs
            else:  # pragma: no cover
                assert len(input_tensor_names) == len(inputs), "inputs len must equal with input_tensor"
                for i, input_tensor_name in enumerate(input_tensor_names):
                    feed_dict[input_tensor_name] = inputs[i]

            _ = infer(**feed_dict)

            if idx >= self.calib_iteration:
                break

    def _check_tf_version(self):  # pragma: no cover
        """Check if the installed tensorflow version is supported."""
        is_supported_version = False
        is_sprbase_version = False
        try:
            from tensorflow import python

            if hasattr(python, "pywrap_tensorflow") and hasattr(python.pywrap_tensorflow, "IsMklEnabled"):
                from tensorflow.python.pywrap_tensorflow import IsMklEnabled
            elif hasattr(python.util, "_pywrap_util_port"):
                from tensorflow.python.util._pywrap_util_port import IsMklEnabled
            else:
                from tensorflow.python._pywrap_util_port import IsMklEnabled
            if IsMklEnabled() and (version1_lte_version2(TF_SUPPORTED_MIN_VERSION, tf.version.VERSION)):
                is_supported_version = True

            if version1_gte_version2(tf.version.VERSION, "2.6.0") and os.getenv("TF_ENABLE_ONEDNN_OPTS") == "1":
                is_supported_version = True

            if version1_gte_version2(tf.version.VERSION, "2.9.0"):
                is_supported_version = True

            if tf.version.VERSION == "1.15.0-up3":  # pragma: no cover
                is_supported_version = True

            if tf.version.VERSION in SPR_BASE_VERSIONS:
                is_supported_version = True
                is_sprbase_version = True

        except Exception as e:
            raise ValueError(e)
        finally:
            if version1_gt_version2(tf.version.VERSION, TF_SUPPORTED_MAX_VERSION) and not is_sprbase_version:
                logger.warning(
                    str(
                        "Please note the {} version of TensorFlow is not fully verified! "
                        "Suggest to use the versions between {} and {} if meet problem."
                    ).format(tf.version.VERSION, TF_SUPPORTED_MIN_VERSION, TF_SUPPORTED_MAX_VERSION)
                )

            if version1_eq_version2(tf.version.VERSION, "2.5.0") and os.getenv("TF_ENABLE_MKL_NATIVE_FORMAT") != "0":
                logger.fatal(
                    "Please set environment variable TF_ENABLE_MKL_NATIVE_FORMAT=0 " "when TensorFlow 2.5.0 installed."
                )

            if (
                version1_gte_version2(tf.version.VERSION, "2.6.0")
                and version1_lt_version2(tf.version.VERSION, "2.9.0")
                and os.getenv("TF_ENABLE_ONEDNN_OPTS") != "1"
            ):
                logger.fatal(
                    "Please set environment variable TF_ENABLE_ONEDNN_OPTS=1 "
                    "when TensorFlow >= 2.6.0 and < 2.9.0 installed."
                )

            if not is_supported_version:
                raise ValueError(
                    str("Please install TensorFlow within version >={} and <={}.").format(
                        TF_SUPPORTED_MIN_VERSION, TF_SUPPORTED_MAX_VERSION
                    )
                )

    def _check_args(self):  # pragma: no cover
        """Check model's arguments."""
        if (
            self.model.workspace_path
            and not os.path.isdir(self.model.workspace_path)
            and not os.path.exists(os.path.dirname(self.model.workspace_path))
        ):
            raise ValueError('"output_graph" directory does not exist.')
        self._output_path = self.model.workspace_path

    def _gen_tmp_filenames(self):
        """Generate the temporary file names."""
        self._int8_dynamic_range_model_path = os.path.join(self._output_path, "int8_dynamic_range_graph")
        self._int8_logged_model_path = os.path.join(self._output_path, "int8_logged_graph")
        self._fp32_logged_model_path = os.path.join(self._output_path, "fp32_logged_graph")
        self._int8_frozen_range_model_path = os.path.join(self._output_path, "int8_frozen_range_graph")
        self._bf16_mixed_precision_model_path = os.path.join(self._output_path, "int8_bf16_mixed_precision_graph")

        self.output_graph = os.path.join(self._output_path, "int8_final_fused_graph")
        if self.performance_only:
            # reuse the fp32 model for performance only mode
            self._tmp_model = self._fp32_model
        else:
            # to keep temp model
            if "backend" in self.model.kwargs:  # pragma: no cover
                self._tmp_model = Model(self.model._model, **self.model.kwargs)
            else:
                self._tmp_model = Model(
                    self.model._model,
                    **self.model.kwargs,
                    backend=(
                        "itex"
                        if self.itex_mode and not isinstance(self.model, TensorflowSavedModelModel)
                        else "default"
                    )
                )
            self._tmp_model.graph_def = self.model.graph_def
            self._tmp_model.output_tensor_names = self.output_tensor_names
            self._tmp_model.input_tensor_names = self.input_tensor_names

    def convert(self):
        """Do conversion.

        Including:
            1) optimize fp32_frozen_graph,
            2) quantize graph,
            3) calibration,
            4) fuse RequantizeOp with fused quantized conv, and so on.
            5) bf16 convert if the self.bf16_ops is not empty

        :return:
        """
        model = self._tmp_model
        if len(self.op_wise_config) > 0:
            if self.qdq_enabled:
                model = self.quantize_with_qdq_pattern()
            else:
                model = self.quantize()

        if self.itex_mode:  # pragma: no cover
            host_const_graph_def = PostHostConstConverter(self._tmp_model.graph_def).do_transformation()
            host_const_graph_def.library.CopyFrom(self.model.graph_def.library)
            self._tmp_model.graph_def = host_const_graph_def

            return self._tmp_model

        if self.exclude_node_names:
            self.bf16_ops.extend(self.exclude_node_names)

        if (
            len(self.bf16_ops) > 0
            and (self.use_bf16 or self.performance_only)
            and (CpuInfo().bf16 or os.getenv("FORCE_BF16") == "1")
        ):
            model = self.bf16_convert()

        if self.new_api:
            if self.performance_only:
                model.graph_def = FuseConvRedundantDequantizeTransformer(model.graph_def).do_transformation()
            post_optimize_graph_def = FuseMatMulRedundantDequantizeTransformer(model.graph_def).do_transformation()
            post_optimize_graph_def.library.CopyFrom(self.model.graph_def.library)
            model.graph_def = post_optimize_graph_def
        post_cse_graph_def = PostCseOptimizer(model.graph_def).do_transformation()
        post_hostconst_graph_def = PostHostConstConverter(post_cse_graph_def).do_transformation()
        post_hostconst_graph_def.library.CopyFrom(self.model.graph_def.library)
        model.graph_def = post_hostconst_graph_def

        if debug:
            model.save(self.output_graph)
            logger.info("Save converted graph file to {}.".format(self.output_graph))
        model.q_config = self.scale_info
        return model

    def _get_fp32_print_node_names(self, specified_op_list):
        """Get the print node name of the fp32 graph."""
        offset_map = {
            "QuantizedConv2DWithBiasSumAndRelu": 3,
            "QuantizedConv2DWithBiasAndRelu": 2,
            "QuantizedConv2DWithBias": 1,
        }
        target_conv_op = []
        sorted_graph = QuantizeGraphHelper().get_sorted_graph(
            self._fp32_model.graph_def, self._fp32_model.input_node_names, self._fp32_model.output_node_names
        )

        node_name_mapping = {node.name: node for node in self._tmp_graph_def.node if node.op != "Const"}

        for node in self._tmp_graph_def.node:
            if node.op in offset_map:
                target_conv_op.append(node.name.split("_eightbit_")[0])
        fp32_node_name_mapping = {node.name: node for node in sorted_graph.node if node.op != "Const"}
        sorted_node_names = [i.name for i in sorted_graph.node if i.op != "Const"]

        output_node_names = []
        for i in target_conv_op:
            if specified_op_list and i not in specified_op_list:
                continue
            if (
                node_name_mapping[i + "_eightbit_quantized_conv"].op == "QuantizedConv2DWithBiasSumAndRelu"
            ):  # pragma: no cover
                start_index = sorted_node_names.index(i)
                for index, value in enumerate(sorted_node_names[start_index:]):
                    if (
                        fp32_node_name_mapping[value].op.startswith("Add")
                        and fp32_node_name_mapping[sorted_node_names[start_index + index + 1]].op == "Relu"
                    ):
                        output_node_names.append(sorted_node_names[start_index + index + 1])
                        self._print_node_mapping[sorted_node_names[start_index + index + 1]] = i

            elif i in sorted_node_names:
                start_index = sorted_node_names.index(i)
                end_index = start_index + offset_map[node_name_mapping[i + "_eightbit_quantized_conv"].op]
                output_node_names.append(sorted_node_names[end_index])
                self._print_node_mapping[sorted_node_names[end_index]] = i

        for i in output_node_names:
            self._kl_keys.append(";" + i + "__print__;__KL")

        fp32_graph_def = graph_pb2.GraphDef()
        fp32_graph_def.CopyFrom(self._fp32_model.graph_def)
        self._fp32_model.graph_def = InsertLogging(
            self._fp32_model.graph_def, node_name_list=output_node_names, message="__KL:", summarize=-1, dump_fp32=True
        ).do_transformation()

        self._fp32_model.save(self._fp32_logged_model_path)
        self._fp32_model.graph_def = fp32_graph_def
        return self._fp32_model

    def _search_y_pattern_for_itex(self):  # pragma: no cover
        """Search the Y pattern for itex and return the op name."""
        g = GraphAnalyzer()
        g.graph = self._fp32_model.graph_def
        g.parse_graph()
        y_pattern = [["Conv2D", "MatMul"], ["BiasAdd"], ["Add", "AddV2", "AddN"], ("Relu",)]
        y_pattern_variant = [["MaxPool", "AvgPool"], ["Add", "AddV2", "AddN"], ("Relu",)]
        target_nodes = g.query_fusion_pattern_nodes(y_pattern)
        target_nodes_variant = g.query_fusion_pattern_nodes(y_pattern_variant)

        res = {}
        for i in target_nodes:
            if i[2] not in res:
                res[i[2]] = 1
            else:
                res[i[2]] += 1
        matched_add_nodes = [(i,) for i in res if res[i] == 2]
        for i in res:
            if res[i] == 1:
                for j in target_nodes_variant:
                    if j[1] == i:
                        matched_add_nodes.append((i,))
        return matched_add_nodes

    def quantize(self):
        """Quantize graph only (without optimizing fp32 graph).

        Including:
            1) quantize graph,
            2) calibration,
            3) fuse RequantizeOp with fused quantized conv, and so on.

        :return:
        """
        try:
            self._quantize_graph()
            self.quantized_node_info = [tuple(i) for i in self.quantized_node_info]

            if self.fake_quant:  # pragma: no cover
                self._fuse_requantize_with_fused_quantized_node()
            else:
                if self._enable_kl_op_names:
                    self._get_fp32_print_node_names(self._enable_kl_op_names)
                    self._generate_calibration_data(self._fp32_logged_model_path, self._fp32_print_data, True)

                output_tensor_names = copy.deepcopy(self.model.output_tensor_names)
                sampling_graph_def = copy.deepcopy(self._fp32_model.graph_def)

                # TODO: this is a workaround to make Min/Max node be completely eliminated in int8 graph
                # after enabling pad+conv2d in new API.
                non_pad_ops = list(list(set(self.fp32_ops).union(set(self.bf16_ops))))
                sampling_graph_def = FusePadWithFP32Conv2DOptimizer(
                    sampling_graph_def, non_pad_ops, self._tmp_model.input_node_names, self.op_wise_config, self.new_api
                ).do_transformation()

                for i in self.quantized_node_info:
                    sampling_graph_def, output_names = InsertPrintMinMaxNode(
                        sampling_graph_def, i[0], i[-1], self.new_api
                    ).do_transformation()
                    output_tensor_names.extend(output_names)
                if self.quantized_node_info:
                    sampling_graph_def.library.CopyFrom(self.model.graph_def.library)
                    self._sampling_model.graph_def = sampling_graph_def
                    self._sampling_model.output_tensor_names = output_tensor_names
                    tmp_dump_file = tempfile.mkstemp(suffix=".log")[1]
                    with CaptureOutputToFile(tmp_dump_file):
                        self._inference(self._sampling_model)
                    self._calibration_data = Helper.gen_valid_sampling_log(tmp_dump_file)

                del output_tensor_names
                del sampling_graph_def
                del self._sampling_model
                import gc

                gc.collect()

                if len(self._calibration_data) > 0:
                    self._freeze_requantization_ranges(self._kl_op_dict)
                    self._fuse_requantize_with_fused_quantized_node()

        except ValueError as e:  # pragma: no cover
            logger.error("Fail to quantize graph due to {}.".format(str(e)))
            self._tmp_model = None
            raise
        except Exception as e:
            import traceback

            traceback.print_exc()
            self._tmp_model = None
            logger.error("Fail to quantize graph due to {}.".format(str(e)))
            return self._tmp_model
        finally:
            if not debug:
                self._post_clean()
        return self._tmp_model

    def bf16_convert(self):
        """Convert fp32 nodes in bf16_node to bf16 dtype based on FP32 + INT8 mixed precision graph."""
        try:
            logger.info("Start BF16 conversion.")
            self._tmp_model.graph_def = BF16Convert(
                self._tmp_model.graph_def, self.fp32_ops, self.bf16_ops
            ).do_transformation()

        except Exception as e:
            import traceback

            traceback.print_exc()
            self._tmp_model = None
            logger.error("Fail to convert graph due to {}.".format(str(e)))
        finally:
            if debug:
                self._tmp_model.save(self._bf16_mixed_precision_model_path)

            return self._tmp_model

    def _quantize_graph(self):
        """Quantize graph."""
        non_pad_ops = list(list(set(self.fp32_ops).union(set(self.bf16_ops))))

        self._tmp_graph_def = FusePadWithConv2DOptimizer(
            self._tmp_graph_def, non_pad_ops, self._tmp_model.input_node_names, self.op_wise_config, self.new_api
        ).do_transformation()

        self._tmp_graph_def = QuantizeGraphHelper().get_sorted_graph(
            self._tmp_graph_def, self._tmp_model.input_node_names, self._tmp_model.output_node_names
        )

        self._tmp_graph_def, self.quantized_node_info, exclude_node_names = QuantizeGraphForIntel(
            self._tmp_graph_def,
            self._tmp_model.input_node_names,
            self._tmp_model.output_node_names,
            self.op_wise_config,
            self.int8_sequences,
            self.device,
            self.fake_quant,
            self.new_api,
            self.performance_only,
            self.itex_mode,
        ).do_transform()
        self.exclude_node_names = exclude_node_names
        self._tmp_graph_def.library.CopyFrom(self.model.graph_def.library)
        if debug and not self.performance_only:
            self._tmp_model.graph_def = self._tmp_graph_def
            self._tmp_model.save(self._int8_dynamic_range_model_path)

    def _generate_calibration_data(self, tmp_path, output_data, enable_kl_algo=False):
        """Generate the calibration data."""
        tmp_dump_file = os.path.join(os.path.dirname(self.output_graph), "requant_min_max.log")

        logger.debug("Generate calibration data and save to {}.".format(tmp_dump_file))

        if "backend" in self._tmp_model.kwargs:
            model = Model(tmp_path, **self._tmp_model.kwargs)
        else:  # pragma: no cover
            model = Model(
                tmp_path,
                **self._tmp_model.kwargs,
                backend=(
                    "itex"
                    if self.itex_mode and not isinstance(self._tmp_model, TensorflowSavedModelModel)
                    else "default"
                )
            )
        model.output_tensor_names = self.output_tensor_names
        model.input_tensor_names = self.input_tensor_names

        with CaptureOutputToFile(tmp_dump_file):
            self._inference(model)

        with open(tmp_dump_file, errors="ignore") as f:
            output_data.extend(f.readlines())

        for line in output_data:
            if enable_kl_algo and line.rsplit(":")[0] in self._kl_keys:
                fp32_data = get_all_fp32_data(line.rsplit(":")[-1])
                key = self._print_node_mapping[line[1:].split("__print")[0]] + "_eightbit_requant_range"
                if key not in self._kl_op_dict:
                    self._kl_op_dict[key] = get_tensor_histogram(fp32_data)
                else:
                    self._kl_op_dict[key] = combine_histogram(self._kl_op_dict[key], fp32_data)

    def _freeze_requantization_ranges(self, additional_data=None):
        """Freeze requantization ranges after doing quantization."""
        self._tmp_graph_def, quantizev2_max = FreezeValueTransformer(
            self._tmp_graph_def, self._calibration_data, "__max:", device=self.device
        ).do_transformation()
        self._tmp_graph_def, quantizev2_min = FreezeValueTransformer(
            self._tmp_graph_def, self._calibration_data, "__min:", device=self.device
        ).do_transformation()
        self._tmp_graph_def, requant_min_max = FreezeValueTransformer(
            self._tmp_graph_def,
            self._calibration_data,
            "__requant_min_max",
            tensor_data=additional_data,
            device=self.device,
        ).do_transformation()

        self.scale_info.update(quantizev2_max)
        self.scale_info.update(quantizev2_min)
        self.scale_info.update(requant_min_max)

        if (
            "scale_propagation_max_pooling" in self.recipes and self.recipes["scale_propagation_max_pooling"]
        ):  # pragma: no cover
            self._tmp_graph_def = ScaleProPagationTransformer(self._tmp_graph_def).do_transformation()

        if debug and not self.new_api:
            self._tmp_graph_def.library.CopyFrom(self.model.graph_def.library)
            self._tmp_model.graph_def = self._tmp_graph_def
            self._tmp_model.save(self._int8_frozen_range_model_path)

    def _fuse_requantize_with_fused_quantized_node(self):
        """Fuse the Requantize/Dequantize with fused quantized Ops."""
        if self.fake_quant:  # pragma: no cover
            self._tmp_graph_def = FreezeFakeQuantOpOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = FuseConvRequantizeTransformer(
            self._tmp_graph_def, self.device, self.new_api
        ).do_transformation()

        if not self.fake_quant:
            if self.qdq_enabled:
                self._tmp_graph_def = FuseMatMulRequantizeNewAPITransformer(self._tmp_graph_def).do_transformation()

                self._tmp_graph_def = FuseMatMulRequantizeDequantizeNewAPITransformer(
                    self._tmp_graph_def
                ).do_transformation()
            else:
                self._tmp_graph_def = FuseMatMulRequantizeTransformer(self._tmp_graph_def).do_transformation()

                self._tmp_graph_def = FuseMatMulRequantizeDequantizeTransformer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = StripUnusedNodesOptimizer(
            self._tmp_graph_def, self._tmp_model.input_node_names, self._tmp_model.output_node_names
        ).do_transformation()

        input_output_names = self._tmp_model.input_node_names + self._tmp_model.output_node_names
        self._tmp_graph_def = RemoveTrainingNodesOptimizer(
            self._tmp_graph_def, protected_nodes=input_output_names
        ).do_transformation()

        self._tmp_graph_def = FoldBatchNormNodesOptimizer(self._tmp_graph_def).do_transformation()

        if self.performance_only or (
            "scale_propagation_concat" in self.recipes and self.recipes["scale_propagation_concat"]
        ):
            self._tmp_graph_def = RerangeQuantizedConcat(
                self._tmp_graph_def, self.device, performance_only=self.performance_only
            ).do_transformation()

        self._tmp_graph_def = MetaInfoChangingMemOpOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = StripEquivalentNodesOptimizer(
            self._tmp_graph_def, self._tmp_model.output_node_names
        ).do_transformation()

        if self.advance_config is not None and deep_get(self.advance_config, "bias_correction") is not None:
            self._tmp_graph_def = BiasCorrection(
                self._tmp_graph_def, self.model.graph_def, self.new_api
            ).do_transformation()

        self._tmp_graph_def.library.CopyFrom(self.model.graph_def.library)

        self._tmp_model.graph_def = self._tmp_graph_def

    def _post_clean(self):  # pragma: no cover
        """Delete the temporarily files generated during the quantization process.

        :return: None
        """
        if os.path.exists(self._int8_logged_model_path) and os.path.isdir(self._int8_logged_model_path):
            import shutil

            shutil.rmtree(self._int8_logged_model_path)

        elif gfile.Exists(self._int8_logged_model_path + ".pb"):
            os.remove(self._int8_logged_model_path + ".pb")

    def quantize_with_qdq_pattern(self):
        """Quantize model by inserting QDQ.

        step 1: insert QDQ pairs and update node info
        step 2: convert Q-DQ-node-Q-DQ to Q-newAPI node-DQ
        """
        try:
            self._insert_qdq_pairs()
            self._convert_qdq()

        except ValueError as e:  # pragma: no cover
            logger.error("Fail to quantize graph due to {}.".format(str(e)))
            self._tmp_model = None
            raise
        except Exception as e:
            import traceback

            traceback.print_exc()
            self._tmp_model = None
            logger.error("Fail to quantize graph due to {}.".format(str(e)))
            return self._tmp_model
        finally:
            if not debug:
                self._post_clean()
        return self._tmp_model

    def _insert_qdq_pairs(self):
        """Insert QDQ pairs before Conv/MatMul/Pooling Ops."""
        # Fuse Pad into Conv2D, Conv3D, DepthwiseConv2dNative
        non_pad_ops = list(list(set(self.fp32_ops).union(set(self.bf16_ops))))
        self._tmp_graph_def = FusePadWithConv2DOptimizer(
            self._tmp_graph_def, non_pad_ops, self._tmp_model.input_node_names, self.op_wise_config, self.new_api, True
        ).do_transformation()

        # Sort graph
        self._tmp_graph_def = QuantizeGraphHelper().get_sorted_graph(
            self._tmp_graph_def, self._tmp_model.input_node_names, self._tmp_model.output_node_names
        )

        self._tmp_graph_def.library.CopyFrom(self.model.graph_def.library)

        # Find out the quantized nodes
        self.quantized_node_info = OptimizeQDQGraph(
            self._tmp_graph_def,
            self._tmp_model.input_node_names,
            self._tmp_model.output_node_names,
            self.op_wise_config,
            self.int8_sequences,
            self.device,
            self.fake_quant,
            self.new_api,
            self.performance_only,
            self.itex_mode,
        ).get_quantized_nodes()

        if self.itex_mode:  # pragma: no cover
            self.quantized_node_info.extend(self._search_y_pattern_for_itex())

        if self._enable_kl_op_names:  # pragma: no cover
            self._get_fp32_print_node_names(self._enable_kl_op_names)
            self._generate_calibration_data(self._fp32_logged_model_path, self._fp32_print_data, True)

        # Calibration using sampling model
        output_tensor_names = copy.deepcopy(self.model.output_tensor_names)
        sampling_graph_def = copy.deepcopy(self._fp32_model.graph_def)
        # TODO: this is a workaround to make Min/Max node be completely eliminated in int8 graph
        # after enabling pad+conv2d in new API.
        non_pad_ops = list(list(set(self.fp32_ops).union(set(self.bf16_ops))))
        sampling_graph_def = FusePadWithFP32Conv2DOptimizer(
            sampling_graph_def, non_pad_ops, self._tmp_model.input_node_names, self.op_wise_config, self.new_api, True
        ).do_transformation()

        for i in self.quantized_node_info:
            sampling_graph_def, output_names = InsertPrintMinMaxNode(
                sampling_graph_def, i[0], i[-1], self.new_api
            ).do_transformation()
            output_tensor_names.extend(output_names)

        if self.quantized_node_info:
            sampling_graph_def.library.CopyFrom(self.model.graph_def.library)
            self._sampling_model.graph_def = sampling_graph_def
            self._sampling_model.output_tensor_names = output_tensor_names
            tmp_dump_file = tempfile.mkstemp(suffix=".log")[1]
            with CaptureOutputToFile(tmp_dump_file):
                self._inference(self._sampling_model)
            self._calibration_data = Helper.gen_valid_sampling_log(tmp_dump_file)

        if hasattr(self._sampling_model, "_weight_tensor_minmax_dict"):
            self._llm_weight_minmax = self._sampling_model.weight_tensor_minmax_dict

        del sampling_graph_def
        del output_tensor_names
        del self._sampling_model
        import gc

        gc.collect()

        # Insert QDQ pattern
        self._tmp_graph_def = GenerateGraphWithQDQPattern(
            self._tmp_graph_def,
            self._calibration_data,
            self.op_wise_config,
            self.fake_quant,
            self.fp32_ops,
            self.bf16_ops,
            self.quantized_node_info,
            self.device,
            self.performance_only,
            self.itex_mode,
            self._llm_weight_minmax,
        ).do_transformation()

    def _convert_qdq(self):
        """Convert Dequantize + Op + QuantizeV2 into QuantizedOps."""
        if self.itex_mode:  # pragma: no cover
            self._tmp_graph_def, quantizev2_max = FreezeValueTransformer(
                self._tmp_graph_def, self._calibration_data, "__max:", self.itex_mode
            ).do_transformation()
            self._tmp_graph_def, quantizev2_min = FreezeValueTransformer(
                self._tmp_graph_def, self._calibration_data, "__min:", self.itex_mode
            ).do_transformation()
            self._tmp_graph_def, requant_min_max = FreezeValueTransformer(
                self._tmp_graph_def,
                self._calibration_data,
                "__requant_min_max",
                tensor_data=self._kl_op_dict,
                device=self.device,
                itex_mode=self.itex_mode,
            ).do_transformation()

            self.scale_info.update(quantizev2_max)
            self.scale_info.update(quantizev2_min)
            self.scale_info.update(requant_min_max)

            self._tmp_graph_def = StripUnusedNodesOptimizer(
                self._tmp_graph_def, self._tmp_model.input_node_names, self._tmp_model.output_node_names
            ).do_transformation()

            self._tmp_graph_def = ShareQDQForItexYPatternOptimizer(self._tmp_graph_def).do_transformation()
            self._tmp_graph_def = MergeDuplicatedQDQOptimizer(self._tmp_graph_def).do_transformation()

            self._tmp_graph_def.library.CopyFrom(self.model.graph_def.library)
            self._tmp_model.graph_def = self._tmp_graph_def
            self._tmp_model.graph_def.library.CopyFrom(self.model.graph_def.library)
        else:
            self._tmp_graph_def, exclude_node_names = OptimizeQDQGraph(
                self._tmp_graph_def,
                self._tmp_model.input_node_names,
                self._tmp_model.output_node_names,
                self.op_wise_config,
                self.int8_sequences,
                self.device,
                self.fake_quant,
                self.new_api,
                self.performance_only,
                self.itex_mode,
            ).do_transform()
            self.exclude_node_names = exclude_node_names

            if len(self._calibration_data) > 0:
                self._freeze_requantization_ranges(self._kl_op_dict)
                self._fuse_requantize_with_fused_quantized_node()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Convert placeholder to const Graph Rewriter."""

from tensorflow.core.framework import attr_value_pb2, node_def_pb2
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class ConvertPlaceholderToConst(GraphRewriterBase):
    """Convert placeholder to const for removing training nodes."""

    @dump_elapsed_time("Pass ConvertPlaceholderToConst")
    def do_transformation(self):
        """Rename the PlaceHolderWithDefault node to constant.

        In a frozen graph, PlaceholderWithDefault nodes can be converted to
        Constant op nodes with same value. This will help simplify the graph.

        Args:
            input_graph_def: A GraphDef containing a model.
            nodes_to_convert: A list of PlaceholderWithDefault or Placeholder
            nodes to be converted to Constants with their new value.

        Returns:
            modified graph with PlaceholderWithDefault node converted to Constant node
        """
        cur_graph = GraphAnalyzer()
        cur_graph.graph = self.model

        graph_info = cur_graph.parse_graph()

        target_nodes = cur_graph.query_fusion_pattern_nodes([["PlaceholderWithDefault"]])
        for i in target_nodes:
            placeholder_node = graph_info[i[0]].node
            new_node = node_def_pb2.NodeDef()
            if dtypes.bool.as_datatype_enum == placeholder_node.attr["dtype"].type:
                placeholder_input_node = None
                if placeholder_node.input:
                    placeholder_input_node = graph_info[Helper.node_name_from_input(placeholder_node.input[0])].node

                if placeholder_input_node and placeholder_input_node.op != "Const":
                    continue
                if placeholder_input_node:
                    new_val_str = placeholder_input_node.attr["value"].tensor.bool_val
                else:
                    continue

                new_node.op = "Const"
                new_node.name = placeholder_node.name + "_const"
                new_node.attr["dtype"].CopyFrom(placeholder_node.attr["dtype"])
                new_node.attr["value"].CopyFrom(
                    attr_value_pb2.AttrValue(
                        tensor=tensor_util.make_tensor_proto(self.strtobool(new_val_str), dtype=dtypes.bool, shape=[])
                    )
                )
                cur_graph.add_node(new_node, None, graph_info[placeholder_node.name].outputs)
                for each_output in graph_info[placeholder_node.name].outputs:
                    for i, input_name in enumerate(graph_info[each_output].node.input):
                        if input_name == placeholder_node.name:
                            new_input = (
                                graph_info[each_output].node.input[:i]
                                + [new_node.name]
                                + graph_info[each_output].node.input[i + 1 :]
                            )
                            graph_info[each_output].node.ClearField("input")
                            graph_info[each_output].node.input.extend(new_input)
                cur_graph.remove_node(placeholder_node.name)
            else:
                continue

        return cur_graph.dump_graph()

    def strtobool(self, val_str):
        """Return boolean value of it's equivalent string representation."""
        if val_str == [True]:
            return True
        if val_str == [False]:
            return False
        return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantize MatMul."""

import numpy as np
from tensorflow.core.framework import graph_pb2, node_def_pb2
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.quantize_graph_common import QuantizeGraphHelper as helper

from .quantize_graph_base import QuantizeNodeBase


class FuseNodeStartWithMatmul(QuantizeNodeBase):
    """Quantize MatMul and apply the fusion."""

    def __init__(self, **kwargs):
        """Initialization."""
        super().__init__(**kwargs)

        self.sorted_patterns = sorted(self.patterns, key=lambda i: len(i), reverse=True)
        self.exclude_matmul_name = []
        self.fusion_op_type = set(fusion[0] for fusion in self.patterns)
        self.fusion_mapping = {
            "MatMulBiasAdd": self.apply_matmul_biasadd_fusion,
            "MatMul": self.apply_matmul_biasadd_fusion,
            "MatMulBiasAddRelu": self.apply_matmul_biasadd_relu_fusion,
            "MatMulRelu": self.apply_matmul_biasadd_relu_fusion,
        }

    def apply_matmul_biasadd_relu_fusion(self, match_node_name):
        """Apply the MatMul BiasAdd Relu fusion."""
        matched_node = self.node_name_mapping[match_node_name[0]]
        control_inputs, normal_inputs = self._get_node_input(matched_node.node.name)
        weight_name = normal_inputs[1]
        weight_node = self.node_name_mapping[helper.node_name_from_input(weight_name)].node

        # FIXME We only quantize the MatMul op which second input node type is const. This is a
        # workaround for RNN model like LTSM.
        if weight_node.op != "Const":
            self.output_graph = self.input_graph
            return []

        weights_content = tensor_util.MakeNdarray(weight_node.attr["value"].tensor)

        if np.any(np.isnan(weights_content)):
            self.output_graph = self.input_graph
            return []

        for i in self.node_name_mapping:
            if weight_node.name in self.node_name_mapping[i].output:
                self.output_graph = self.input_graph
                return []
        second_node = self.node_name_mapping[match_node_name[1]].node
        skip_node_name = match_node_name[1:]

        need_insert_dummy_biasadd = 1
        offset = 1
        if len(match_node_name) == 3:
            add_a_node_name = helper.node_name_from_input(second_node.input[0])
            add_a_node = self.node_name_mapping[add_a_node_name].node
            add_b_node_name = helper.node_name_from_input(second_node.input[1])
            add_b_node = self.node_name_mapping[add_b_node_name].node
            if add_a_node.op != "Const" and add_b_node.op == "Const":
                need_insert_dummy_biasadd = 0
                offset = 0
            if need_insert_dummy_biasadd:
                self.apply_matmul_biasadd_fusion(match_node_name[:1])
                return match_node_name[:1]

        q_weights_name, q_weights_min_name, q_weights_max_name = self._intel_cpu_quantize_weight_eightbit(
            matched_node.node.op, self.node_name_mapping[weight_name].node, self.per_channel
        )

        skip_node_name.append(weight_name)

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                pass
            elif node.name == match_node_name[0]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))

                quantized_node_name = node.name + "_eightbit_quantized_mat_mul"
                if need_insert_dummy_biasadd:
                    t_b_index = 0 if matched_node.node.attr["transpose_b"].b else 1
                    bias_size = weights_content.shape[t_b_index]
                    bias_node_name = node.name + "_fake_bias"
                    bias_node = helper.create_constant_node(
                        bias_node_name, [0] * bias_size, dtypes.float32, shape=[bias_size]
                    )
                    self.add_output_graph_node(bias_node)
                else:
                    bias_node_name = self.node_name_mapping[match_node_name[1]].node.input[1]

                relu_node_name = match_node_name[2 - offset]
                all_input_names = self._add_eightbit_prologue_nodes(matched_node.node.name)
                all_input_names = all_input_names[:1] + [q_weights_name] + all_input_names[1:]
                all_input_names.append(q_weights_min_name)
                all_input_names.append(q_weights_max_name)
                quantized_node_input_names = (
                    all_input_names[:2] + [bias_node_name] + all_input_names[2:] + control_inputs
                )

                quantized_matmul_node = helper.create_node(
                    "QuantizedMatMulWithBiasAndRelu", quantized_node_name, quantized_node_input_names
                )

                helper.copy_attr(quantized_matmul_node, "transpose_a", node.attr["transpose_a"])
                helper.copy_attr(quantized_matmul_node, "transpose_b", node.attr["transpose_b"])
                helper.set_attr_dtype(quantized_matmul_node, "T1", dtypes.quint8)
                helper.set_attr_dtype(quantized_matmul_node, "T2", dtypes.qint8)
                helper.set_attr_dtype(quantized_matmul_node, "Toutput", dtypes.qint32)
                helper.set_attr_string(
                    quantized_matmul_node, "input_quant_mode", b"MIN_FIRST" if self.is_asymmetric else b"SCALED"
                )

                self.add_output_graph_node(quantized_matmul_node)

                quantize_down_name = self._add_quantize_down_nodes(node, quantized_node_name, dtypes.quint8, False)
                self._intel_cpu_add_dequantize_result_node(quantize_down_name, relu_node_name)
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)
        return match_node_name

    def apply_matmul_biasadd_fusion(self, match_node_name):
        """Apply MatMul BiasAdd fusion."""
        skip_node_name = match_node_name[1:]
        matched_node = self.node_name_mapping[match_node_name[0]]
        control_inputs, normal_inputs = self._get_node_input(matched_node.node.name)
        weight_name = normal_inputs[1]
        weight_node = self.node_name_mapping[helper.node_name_from_input(weight_name)].node

        enter_node = None

        if weight_node.op == "Enter":
            parent_node = self.node_name_mapping[helper.node_name_from_input(weight_node.input[0])].node
            # FIXME We only quantize the MatMul op which second input node type is const. This is a
            # workaround for RNN model like LTSM.
            if parent_node.op != "Const":
                self.logger.debug("The weight node of matched_node {} is not Const or Const + Enter, skipped")
                self.output_graph = self.input_graph
                return []
            enter_node = weight_node
            weight_node = parent_node
            weight_name = weight_node.name

        if weight_node.op != "Const":
            self.output_graph = self.input_graph
            return []

        # TODO Remove below two lines once the TF enabled the old QuantizedMatMul while
        # transpose_a/transpose_b could be set to True.
        if matched_node.node.attr["transpose_a"].b is True or matched_node.node.attr["transpose_b"].b is True:
            self.exclude_matmul_name.append(match_node_name[0])
            self.output_graph = self.input_graph
            return []

        if weight_node.op == "Const":
            weights_content = tensor_util.MakeNdarray(weight_node.attr["value"].tensor)

            if np.any(np.isnan(weights_content)):
                self.output_graph = self.input_graph
                return []

        for i in self.node_name_mapping:
            if (
                weight_node.input
                and not weight_node.input[0].startswith("^")
                and weight_node.name in self.node_name_mapping[i].output
            ):
                self.output_graph = self.input_graph
                return []

        len_output = len(matched_node.output)
        is_shared_output = False
        if len_output == 2:
            if (
                self.node_name_mapping[matched_node.output[0]].node.op == "Reshape"
                or self.node_name_mapping[matched_node.output[1]].node.op == "Reshape"
            ):
                is_shared_output = False
            else:
                is_shared_output = True
        elif len_output > 1:
            is_shared_output = True

        need_insert_dummy_biasadd = 1
        if len(match_node_name) == 1:
            if is_shared_output:
                self.output_graph = self.input_graph
                return []
        else:
            second_node = self.node_name_mapping[match_node_name[1]].node
            add_a_node_name = helper.node_name_from_input(second_node.input[0])
            add_a_node = self.node_name_mapping[add_a_node_name].node
            add_b_node_name = helper.node_name_from_input(second_node.input[1])
            add_b_node = self.node_name_mapping[add_b_node_name].node
            if add_a_node.op != "Const" and add_b_node.op in ("Const", "Enter"):
                need_insert_dummy_biasadd = 0
            if need_insert_dummy_biasadd:
                self.apply_matmul_biasadd_fusion(match_node_name[:1])
                return match_node_name[:1]

        if self.frame_info and not enter_node:
            from collections import OrderedDict

            frame_info = OrderedDict(self.frame_info)
            if match_node_name[0] in frame_info and frame_info[match_node_name[0]]:
                enter_node = helper.create_node("Enter", weight_name + "_enter", [weight_name])
                helper.set_attr_string(enter_node, "frame_name", frame_info[weight_name].attr["frame_name"].s)
                helper.set_attr_dtype(enter_node, "T", dtypes.float32)
                helper.set_attr_bool(enter_node, "is_constant", True)
                helper.set_attr_int(
                    enter_node, "parallel_iterations", frame_info[weight_name].attr["parallel_iterations"].i
                )

        q_weights_name, q_weights_min_name, q_weights_max_name = self._intel_cpu_quantize_weight_eightbit(
            matched_node.node.op, self.node_name_mapping[weight_name].node, self.per_channel, enter_node
        )

        skip_node_name.append(weight_name)
        if enter_node:
            skip_node_name.append(enter_node.name)

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                pass
            elif node.name == match_node_name[0]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))

                quantized_node_name = node.name + "_eightbit_quantized_mat_mul"

                if need_insert_dummy_biasadd:
                    t_b_index = 0 if matched_node.node.attr["transpose_b"].b else 1
                    bias_size = weights_content.shape[t_b_index]
                    bias_node_name = node.name + "_fake_bias"
                    bias_node = helper.create_constant_node(
                        bias_node_name, [0] * bias_size, dtypes.float32, shape=[bias_size]
                    )
                    if enter_node:
                        bias_enter_node = helper.create_node("Enter", bias_node_name + "_enter", [bias_node_name])
                        helper.set_attr_string(bias_enter_node, "frame_name", enter_node.attr["frame_name"].s)
                        helper.set_attr_dtype(bias_enter_node, "T", dtypes.float32)
                        helper.set_attr_bool(bias_enter_node, "is_constant", True)
                        helper.set_attr_int(
                            bias_enter_node, "parallel_iterations", enter_node.attr["parallel_iterations"].i
                        )

                        self.add_output_graph_node(bias_enter_node)
                        bias_node_name = bias_enter_node.name

                    self.add_output_graph_node(bias_node)
                else:
                    bias_node_name = self.node_name_mapping[match_node_name[1]].node.input[1]
                    if self.node_name_mapping[bias_node_name].node.op == "Enter":
                        bias_enter_node = helper.create_node("Enter", bias_node_name + "_enter", [bias_node_name])
                        helper.set_attr_string(
                            bias_enter_node,
                            "frame_name",
                            self.node_name_mapping[bias_node_name].node.attr["frame_name"].s,
                        )
                        helper.set_attr_dtype(bias_enter_node, "T", dtypes.float32)
                        helper.set_attr_bool(bias_enter_node, "is_constant", True)
                        helper.set_attr_int(
                            bias_enter_node,
                            "parallel_iterations",
                            self.node_name_mapping[bias_node_name].node.attr["parallel_iterations"].i,
                        )
                        self.add_output_graph_node(bias_enter_node)

                all_input_names = self._add_eightbit_prologue_nodes(matched_node.node.name)
                all_input_names = all_input_names[:1] + [q_weights_name] + all_input_names[1:]
                all_input_names.append(q_weights_min_name)
                all_input_names.append(q_weights_max_name)
                quantized_node_input_names = (
                    all_input_names[:2] + [bias_node_name] + all_input_names[2:] + control_inputs
                )

                quantized_matmul_node = helper.create_node(
                    "QuantizedMatMulWithBias", quantized_node_name, quantized_node_input_names
                )

                helper.copy_attr(quantized_matmul_node, "transpose_a", node.attr["transpose_a"])
                helper.copy_attr(quantized_matmul_node, "transpose_b", node.attr["transpose_b"])
                helper.set_attr_dtype(quantized_matmul_node, "T1", dtypes.quint8)
                helper.set_attr_dtype(quantized_matmul_node, "T2", dtypes.qint8)
                helper.set_attr_dtype(quantized_matmul_node, "Toutput", dtypes.qint32)
                helper.set_attr_dtype(quantized_matmul_node, "Tbias", dtypes.float32)
                helper.set_attr_string(
                    quantized_matmul_node, "input_quant_mode", b"MIN_FIRST" if self.is_asymmetric else b"SCALED"
                )

                self.add_output_graph_node(quantized_matmul_node)
                requantize_type = dtypes.qint8

                quantize_down_name = self._add_quantize_down_nodes(node, quantized_node_name, requantize_type, False)
                self._intel_cpu_add_dequantize_result_node(
                    quantize_down_name,
                    match_node_name[0] if need_insert_dummy_biasadd else match_node_name[1],
                    requantize_type,
                )
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)
        return match_node_name

    def get_longest_fuse(self):
        """Get the longest fusion pattern."""
        self._get_op_list()
        matched_rule, _ = self._is_match(self.sorted_patterns)
        return matched_rule

    def apply_the_transform(self):
        """Quantize MatMul and apply the fusion pattern."""
        self._get_op_list()
        matched_rule, matched_node_name = self._is_match(self.sorted_patterns)

        if matched_node_name:
            self.output_graph = graph_pb2.GraphDef()
            fusion_name = "".join(matched_rule)
            if fusion_name in self.fusion_mapping:
                matched_nodes = self.fusion_mapping[fusion_name](matched_node_name)
            else:  # pragma: no cover
                self.logger.debug("Unknown fusion pattern {}.".format(fusion_name))
                if self.remove_redundant_quant_flag:
                    self.input_graph = self.remove_redundant_quantization(self.input_graph)
                return self.input_graph, []
            self.input_graph = self.output_graph
            self._reset_output_node_maps()
            if self.remove_redundant_quant_flag:
                self.output_graph = self.remove_redundant_quantization(self.output_graph)
            return self.output_graph, matched_nodes, self.exclude_matmul_name

        return self.input_graph, [], []
